#
#  Copyright (c) 2011-2014 Exxeleron GmbH
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import socket
import struct
import threading
import time

from .qmetaconversion import MetaData, CONVERSION_OPTIONS

# Debug logging control
DEBUG = False

from .qtype import QException
from .qreader import QReader, QReaderException
from .qwriter import QWriter, QWriterException



class QConnectionException(Exception):
    '''Raised when a connection to the q service cannot be established.'''
    pass



class QAuthenticationException(QConnectionException):
    '''Raised when a connection to the q service is denied.'''
    pass



class MessageType(object):
    '''Enumeration defining IPC protocol message types.'''
    ASYNC = 0
    SYNC = 1
    RESPONSE = 2

class ListenerThread(threading.Thread):
    def __init__(self, connection, callback_handler=None, data_only=False, raw=False):
        super(ListenerThread, self).__init__()
        self.connection = connection
        self.callback_handler = callback_handler
        self.messages = []
        self._stopper = threading.Event()
        self.data_only = data_only
        self.raw = raw

    def stop(self):
        self._stopper.set()

    def stopped(self):
        return self._stopper.is_set()

    def is_alive(self):
        """Check if the listener thread is still running"""
        return threading.Thread.is_alive(self)

    def wait_for_response(self, timeout_seconds):
        """
        Clear messages and wait for a response message (type 2).

        Args:
            timeout_seconds: Maximum time to wait for response

        Returns:
            bool - True if response arrived within timeout, False otherwise
        """
        self.messages.clear()  # Start fresh
        start_time = time.time()

        while time.time() - start_time < timeout_seconds:
            # If thread is dead, return False immediately
            if not self.is_alive():
                return False

            if self.messages and self.messages[-1].type == MessageType.RESPONSE:
                return True
            time.sleep(0.01)  # 10ms polling

        return False

    def run(self):
        import logging
        import traceback
        import os

        # Set up logging only if DEBUG is enabled
        logger = None
        if DEBUG:
            log_file = '/tmp/qmcp_listener.log'
            logger = logging.getLogger('ListenerThread')

            # Remove any existing handlers to avoid duplicates
            for handler in logger.handlers[:]:
                logger.removeHandler(handler)

            # Create file handler
            file_handler = logging.FileHandler(log_file, mode='a')
            file_handler.setLevel(logging.DEBUG)

            # Create formatter
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            file_handler.setFormatter(formatter)

            # Add handler to logger
            logger.addHandler(file_handler)
            logger.setLevel(logging.DEBUG)

            logger.info("ListenerThread started")

        while not self.stopped():
            try:
                message = self.connection.receive(data_only=self.data_only, raw=self.raw)
                if self.callback_handler:
                    self.callback_handler(message)
                else:
                    self.messages.append(message)
            except QException as e:
                if DEBUG and logger:
                    logger.error(f"Q exception in listener: {e}")
                    logger.debug(f"Q exception traceback: {traceback.format_exc()}")
            except Exception as e:
                if DEBUG and logger:
                    logger.error(f"Socket/connection error in listener: {e}")
                    logger.debug(f"Full traceback: {traceback.format_exc()}")
                    logger.info("Stopping listener thread due to connection error")
                # Connection is likely dead, stop the thread
                break


class QConnection(object):
    '''Connector class for interfacing with the q service.
    
    Provides methods for synchronous and asynchronous interaction.
    
    The :class:`.QConnection` class provides a context manager API and can be 
    used with a ``with`` statement::
    
        with qconnection.QConnection(host = 'localhost', port = 5000) as q:
            print(q)
            print(q('{`int$ til x}', 10))
    
    :Parameters:
     - `host` (`string`) - q service hostname
     - `port` (`integer`) - q service port
     - `username` (`string` or `None`) - username for q authentication/authorization
     - `password` (`string` or `None`) - password for q authentication/authorization
     - `timeout` (`nonnegative float` or `None`) - set a timeout on blocking socket operations
     - `encoding` (`string`) - string encoding for data deserialization
     - `reader_class` (subclass of `QReader`) - data deserializer
     - `writer_class` (subclass of `QWriter`) - data serializer
    :Options: 
     - `raw` (`boolean`) - if ``True`` returns raw data chunk instead of parsed 
       data, **Default**: ``False``
     - `numpy_temporals` (`boolean`) - if ``False`` temporal vectors are
       backed by raw q representation (:class:`.QTemporalList`, 
       :class:`.QTemporal`) instances, otherwise are represented as 
       `numpy datetime64`/`timedelta64` arrays and atoms,
       **Default**: ``False``
     - `single_char_strings` (`boolean`) - if ``True`` single char Python 
       strings are encoded as q strings instead of chars, **Default**: ``False``
    '''

    MAX_PROTOCOL_VERSION = 6

    def __init__(self, host, port, username = None, password = None, timeout = None, encoding = 'latin-1', reader_class = None, writer_class = None, **options):
        self.host = host
        self.port = port
        self.username = username
        self.password = password

        self._connection = None
        self._connection_file = None
        self._protocol_version = None

        self.timeout = timeout

        self._encoding = encoding

        self._options = MetaData(**CONVERSION_OPTIONS.union_dict(**options))

        # Always use base QReader (pandas support is now conditional)
        self._reader_class = QReader
        
        try:
            from ._pandas import PandasQWriter
            self._writer_class = PandasQWriter
        except (ImportError, AssertionError):
            self._writer_class = QWriter

        if reader_class:
            self._reader_class = reader_class

        if writer_class:
            self._writer_class = writer_class


    def __enter__(self):
        self.open()
        return self


    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


    @property
    def protocol_version(self):
        '''Retrieves established version of the IPC protocol.
        
        :returns: `integer` -- version of the IPC protocol
        '''
        return self._protocol_version


    def open(self):
        '''Initialises connection to q service.
        
        If the connection hasn't been initialised yet, invoking the 
        :func:`.open` creates a new socket and performs a handshake with a q 
        service.
        
        :raises: :class:`.QConnectionException`, :class:`.QAuthenticationException` 
        '''
        if not self._connection:
            if not self.host:
                raise QConnectionException('Host cannot be None')

            self._init_socket()
            self._initialize()

            self._writer = self._writer_class(self._connection, protocol_version = self._protocol_version, encoding = self._encoding)
            self._reader = self._reader_class(self._connection_file, encoding = self._encoding)


    def _init_socket(self):
        '''Initialises the socket used for communicating with a q service,'''
        try:
            self._connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._connection.connect((self.host, self.port))
            self._connection.settimeout(self.timeout)
            self._connection_file = self._connection.makefile('b')
        except:
            self._connection = None
            self._connection_file = None
            raise


    def close(self):
        '''Closes connection with the q service.'''
        if self._connection:
            self._connection_file.close()
            self._connection_file = None
            self._connection.close()
            self._connection = None


    def is_connected(self):
        '''Checks whether connection with a q service has been established. 
        
        Connection is considered inactive when: 
         - it has not been initialised, 
         - it has been closed.
         
        :returns: `boolean` -- ``True`` if connection has been established, 
                  ``False`` otherwise
        '''
        return True if self._connection else False


    def _initialize(self):
        '''Performs a IPC protocol handshake.'''
        credentials = (self.username if self.username else '') + ':' + (self.password if self.password else '')
        credentials = credentials.encode(self._encoding)
        self._connection.send(credentials + bytes([self.MAX_PROTOCOL_VERSION, 0]))
        response = self._connection.recv(1)

        if len(response) != 1:
            self.close()
            self._init_socket()

            self._connection.send(credentials + b'\0')
            response = self._connection.recv(1)
            if len(response) != 1:
                self.close()
                raise QAuthenticationException('Connection denied.')

        self._protocol_version = min(struct.unpack('B', response)[0], self.MAX_PROTOCOL_VERSION)


    def __str__(self):
        return '%s@:%s:%s' % (self.username, self.host, self.port) if self.username else ':%s:%s' % (self.host, self.port)


    def query(self, msg_type, query, *parameters, **options):
        '''Performs a query against a q service.
        
        In typical use case, `query` is the name of the function to call and 
        `parameters` are its parameters. When `parameters` list is empty, the 
        query can be an arbitrary q expression (e.g. ``0 +/ til 100``).
        
        Calls a anonymous function with a single parameter:
        
            >>> q.query(qconnection.MessageType.SYNC,'{til x}', 10)
        
        Executes a q expression:
        
            >>> q.query(qconnection.MessageType.SYNC,'til 10')
        
        :Parameters:
         - `msg_type` (one of the constants defined in :class:`.MessageType`) - 
           type of the query to be executed
         - `query` (`string`) - query to be executed
         - `parameters` (`list` or `None`) - parameters for the query
        :Options:
         - `single_char_strings` (`boolean`) - if ``True`` single char Python 
           strings are encoded as q strings instead of chars, 
           **Default**: ``False``
        
        :raises: :class:`.QConnectionException`, :class:`.QWriterException`
        '''
        if not self._connection:
            raise QConnectionException('Connection is not established.')

        if parameters and len(parameters) > 8:
            raise QWriterException('Too many parameters.')

        if not parameters or len(parameters) == 0:
            self._writer.write(query, msg_type, **self._options.union_dict(**options))
        else:
            self._writer.write([query] + list(parameters), msg_type, **self._options.union_dict(**options))


    def sendSync(self, query, *parameters, **options):
        '''Performs a synchronous query against a q service and returns parsed
        data.

        In typical use case, `query` is the name of the function to call and
        `parameters` are its parameters. When `parameters` list is empty, the
        query can be an arbitrary q expression (e.g. ``0 +/ til 100``).

        Executes a q expression:

            >>> print(q.sendSync('til 10'))
            [0 1 2 3 4 5 6 7 8 9]

        Executes an anonymous q function with a single parameter:

            >>> print(q.sendSync('{til x}', 10))
            [0 1 2 3 4 5 6 7 8 9]

        Executes an anonymous q function with two parameters:

            >>> print(q.sendSync('{y + til x}', 10, 1))
            [ 1  2  3  4  5  6  7  8  9 10]

            >>> print(q.sendSync('{y + til x}', *[10, 1]))
            [ 1  2  3  4  5  6  7  8  9 10]

        The :func:`.sendSync` is called from the overloaded :func:`.__call__`
        function. This allows :class:`.QConnection` instance to be called as
        a function:

            >>> print(q('{y + til x}', 10, 1))
            [ 1  2  3  4  5  6  7  8  9 10]


        :Parameters:
         - `query` (`string`) - query to be executed
         - `parameters` (`list` or `None`) - parameters for the query
        :Options:
         - `raw` (`boolean` or `None`) - if ``True`` returns raw data chunk instead of
           parsed data, if ``False`` returns parsed data, if ``None`` uses connection default.
           When set to a boolean, updates the connection's raw setting for this and future queries.
           **Default**: ``None``
         - `numpy_temporals` (`boolean`) - if ``False`` temporal vectors are
           backed by raw q representation (:class:`.QTemporalList`,
           :class:`.QTemporal`) instances, otherwise are represented as
           `numpy datetime64`/`timedelta64` arrays and atoms,
           **Default**: ``False``
         - `single_char_strings` (`boolean`) - if ``True`` single char Python
           strings are encoded as q strings instead of chars,
           **Default**: ``False``

        :returns: query result parsed to Python data structures

        :raises: :class:`.QConnectionException`, :class:`.QWriterException`,
                 :class:`.QReaderException`
        '''
        # Handle raw parameter specially - update connection options if specified
        if 'raw' in options and options['raw'] is not None:
            raw_value = options['raw']
            if isinstance(raw_value, bool):
                # Update connection options for this and future queries
                self._options.raw = raw_value
                if hasattr(self, '_reader') and self._reader and hasattr(self._reader, '_options'):
                    self._reader._options.raw = raw_value

        self.query(MessageType.SYNC, query, *parameters, **options)
        response = self.receive(data_only = False, **options)

        if response.type == MessageType.RESPONSE:
            return response.data
        else:
            self._writer.write(QException('nyi: qPython expected response message'), MessageType.ASYNC if response.type == MessageType.ASYNC else MessageType.RESPONSE)
            raise QReaderException('Received message of type: %s where response was expected')

    def sendSyncDetached(self, query, *parameters, **options):
        '''Sends a synchronous query without waiting for the response.

        Unlike sendAsync(), this generates a response from q which can be
        caught by a listener thread, along with any (neg .z.w) callbacks.

        :Parameters:
         - `query` (`string`) - query to be executed
         - `parameters` (`list` or `None`) - parameters for the query
        :Options:
         - `raw` (`boolean` or `None`) - if specified, updates the connection's raw setting
           for this and future queries. **Default**: ``None``
        '''
        # Handle raw parameter specially - update connection options if specified
        if 'raw' in options and options['raw'] is not None:
            raw_value = options['raw']
            if isinstance(raw_value, bool):
                # Update connection options for this and future queries
                self._options.raw = raw_value
                if hasattr(self, '_reader') and self._reader and hasattr(self._reader, '_options'):
                    self._reader._options.raw = raw_value

        self.query(MessageType.SYNC, query, *parameters, **options)
        
    def sendAsync(self, query, *parameters, **options):
        '''Performs an asynchronous query and returns **without** retrieving of
        the response.

        In typical use case, `query` is the name of the function to call and
        `parameters` are its parameters. When `parameters` list is empty, the
        query can be an arbitrary q expression (e.g. ``0 +/ til 100``).

        Calls a anonymous function with a single parameter:

            >>> q.sendAsync('{til x}', 10)

        Executes a q expression:

            >>> q.sendAsync('til 10')

        :Parameters:
         - `query` (`string`) - query to be executed
         - `parameters` (`list` or `None`) - parameters for the query
        :Options:
         - `raw` (`boolean` or `None`) - if specified, updates the connection's raw setting
           for this and future queries. **Default**: ``None``
         - `single_char_strings` (`boolean`) - if ``True`` single char Python
           strings are encoded as q strings instead of chars,
           **Default**: ``False``

        :raises: :class:`.QConnectionException`, :class:`.QWriterException`
        '''
        # Handle raw parameter specially - update connection options if specified
        if 'raw' in options and options['raw'] is not None:
            raw_value = options['raw']
            if isinstance(raw_value, bool):
                # Update connection options for this and future queries
                self._options.raw = raw_value
                if hasattr(self, '_reader') and self._reader and hasattr(self._reader, '_options'):
                    self._reader._options.raw = raw_value

        self.query(MessageType.ASYNC, query, *parameters, **options)


    def receive(self, data_only = True, **options):
        '''Reads and (optionally) parses the response from a q service.
        
        Retrieves query result along with meta-information:
        
            >>> q.query(qconnection.MessageType.SYNC,'{x}', 10)
            >>> print(q.receive(data_only = False, raw = False))
            QMessage: message type: 2, data size: 13, is_compressed: False, data: 10

        Retrieves parsed query result:

            >>> q.query(qconnection.MessageType.SYNC,'{x}', 10)
            >>> print(q.receive(data_only = True, raw = False))
            10

        Retrieves not-parsed (raw) query result:
        
            >>> from binascii import hexlify
            >>> q.query(qconnection.MessageType.SYNC,'{x}', 10)
            >>> print(hexlify(q.receive(data_only = True, raw = True)))
            fa0a000000
                
        :Parameters:
         - `data_only` (`boolean`) - if ``True`` returns only data part of the 
           message, otherwise returns data and message meta-information 
           encapsulated in :class:`.QMessage` instance 
        :Options:
         - `raw` (`boolean`) - if ``True`` returns raw data chunk instead of 
           parsed data, **Default**: ``False``
         - `numpy_temporals` (`boolean`) - if ``False`` temporal vectors are
           backed by raw q representation (:class:`.QTemporalList`, 
           :class:`.QTemporal`) instances, otherwise are represented as 
           `numpy datetime64`/`timedelta64` arrays and atoms,
           **Default**: ``False``
        
        :returns: depending on parameter flags: :class:`.QMessage` instance, 
                  parsed message, raw data 
        :raises: :class:`.QReaderException`
        '''
        result = self._reader.read(**self._options.union_dict(**options))
        return result.data if data_only else result


    def __call__(self, *parameters, **options):
        return self.sendSync(parameters[0], *parameters[1:], **options)

    def _safe_decode(self, data):
        """
        Safely decode bytes to string with UTF-8/Latin-1 fallback
        
        :returns: decoded string
        """
        if isinstance(data, bytes):
            try:
                return data.decode('utf-8')
            except UnicodeDecodeError:
                return data.decode('latin-1')  # Never fails
        else:
            return str(data)

    def query_with_console(self, *parameters, **options):
        """
        Execute query and return both console output and raw data.
        
        Prefixes the first parameter with {(.Q.s x; x)} to get both
        formatted console output and the actual data structure.
        
        :returns: tuple of (console_output_string, raw_data)
        """
        if not parameters:
            raise ValueError("query_with_console requires at least one parameter")
        
        try:
            # Clear any existing logs before running the query
            try:
                self.sendSync(".qython.delete_logs[]")
            except:
                # Silently ignore if .qython.delete_logs[] doesn't exist
                pass
            
            # Prefix the first parameter with the console output wrapper
            q_prefix = '''{v:$[`trp in key .Q; .Q.trp[{( (1b;`) ;value x)};x;{((0b;`);x;$[4<count y; .Q.sbt -4 _ y; \"\"])}]; ((1b;`);value x)]; a:99999>@[-22!;v;{0}]; (a;$[a;v;0b];.Q.s v 1)}'''
            #prefixed_query = f"{q_prefix} {parameters[0]}"
            #result = self.sendSync(prefixed_query, *parameters[1:], **options)
            result = self.sendSync(q_prefix, parameters[0])
            result = [str.__str__(result[2]), result[1][1]]
            # Handle None result or non-iterable result (like QLambda) - occurs with empty queries
            if result is None or not hasattr(result, '__iter__') or isinstance(result, str):
                return ("", result)
            # Handle None result[0] safely
            # Safely decode the console output
            return result
            
        except QException as e:
            # Q error occurred - extract and decode the actual error content
            if e.args and len(e.args) > 0:
                error_content = e.args[0]  # Extract the actual error bytes/content
                error_str = self._safe_decode(error_content)
            else:
                error_str = self._safe_decode(str(e))
            raise QException(f"Q Error: {error_str}")
        except Exception as e:
            # Print full traceback for debugging
            import traceback
            print(f"Exception in query_with_console: {e}")
            traceback.print_exc()
            raise


def parse_raw_bytes(raw_bytes):
    """
    Parse raw q bytes to Python structures with pandas support.

    Args:
        raw_bytes: Raw bytes from q server

    Returns:
        Parsed Python data structure
    """
    from .qreader import QReader
    from .qmetaconversion import MetaData, CONVERSION_OPTIONS
    from ._pandas import read_object_pandas

    # Create a temporary reader for parsing
    parser = QReader(None, encoding='latin-1')
    parser._options = MetaData(**CONVERSION_OPTIONS.union_dict(pandas=True))
    parser._is_native = True

    # Parse the raw bytes
    return read_object_pandas(parser, raw_bytes)



#
#  Copyright (c) 2011-2014 Exxeleron GmbH
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import struct
import sys
if sys.version > '3':
    from sys import intern
    unicode = str

from .qmetaconversion import MetaData, CONVERSION_OPTIONS
from .qtype import * 
from .qcollection import qlist, QDictionary, qtable, QTable, QKeyedTable
from .qtemporal import qtemporal, from_raw_qtemporal, array_from_raw_qtemporal
import numpy

try:
    from .fastutils import uncompress
except:
    from .utils import uncompress

class QReaderException(Exception):
    '''
    Indicates an error raised during data deserialization.
    '''
    pass



class QMessage(object):
    '''
    Represents a single message parsed from q protocol.
    Encapsulates data, message size, type, compression flag.

    :Parameters:
     - `data` - data payload
     - `message_type` (one of the constants defined in :class:`.MessageType`) -
       type of the message
     - `message_size` (`int`) - size of the message
     - `compression_mode` (`int`) - indicates whether message is compressed, 1 for size<2GB, 2 for larger
    '''

    @property
    def data(self):
        '''Parsed data.'''
        return self._data

    @data.setter
    def data(self, value):
        self._data = value


    @property
    def type(self):
        '''Type of the message.'''
        return self._type


    @property
    def compression_mode(self):
        '''Indicates whether source message was compressed.'''
        return self._compression_mode


    @property
    def size(self):
        '''Size of the source message.'''
        return self._size


    def __init__(self, data, message_type, message_size, compression_mode):
        self._data = data
        self._type = message_type
        self._size = message_size
        self._compression_mode = compression_mode


    def __str__(self, *args, **kwargs):
        return 'QMessage: message type: %s, data size: %s, compression_mode: %s, data: %s' % (self._type, self._size, self._compression_mode, self._data)



class QReader(object):
    '''
    Provides deserialization from q IPC protocol.

    :Parameters:
     - `stream` (`file object` or `None`) - data input stream
     - `encoding` (`string`) - encoding for characters parsing

    :Attrbutes:
     - `_reader_map` - stores mapping between q types and functions
       responsible for parsing into Python objects
    '''

    _reader_map = {}
    parse = Mapper(_reader_map)


    def __init__(self, stream, encoding = 'latin-1'):
        self._stream = stream
        self._buffer = QReader.BytesBuffer()
        self._encoding = encoding


    def read(self, source = None, **options):
        '''
        Reads and optionally parses a single message.

        :Parameters:
         - `source` - optional data buffer to be read, if not specified data is
           read from the wrapped stream
        :Options:
         - `raw` (`boolean`) - indicates whether read data should parsed or
           returned in raw byte form
         - `numpy_temporals` (`boolean`) - if ``False`` temporal vectors are
           backed by raw q representation (:class:`.QTemporalList`,
           :class:`.QTemporal`) instances, otherwise are represented as
           `numpy datetime64`/`timedelta64` arrays and atoms,
           **Default**: ``False``

        :returns: :class:`.QMessage` - read data (parsed or raw byte form) along
                  with meta information
        '''
        message = self.read_header(source)
        message.data = self.read_data(message.size, message.compression_mode, **options)

        return message


    def read_header(self, source = None):
        '''
        Reads and parses message header.

        .. note:: :func:`.read_header` wraps data for further reading in internal
                  buffer

        :Parameters:
         - `source` - optional data buffer to be read, if not specified data is
           read from the wrapped stream

        :returns: :class:`.QMessage` - read meta information
        '''
        if self._stream:
            header = self._read_bytes(8)
            self._buffer.wrap(header)
        else:
            self._buffer.wrap(source)

        self._buffer.endianness = '<' if self._buffer.get_byte() == 1 else '>'
        self._is_native = self._buffer.endianness == ('<' if sys.byteorder == 'little' else '>')
        message_type = self._buffer.get_byte()
        message_compression_mode = self._buffer.get_byte()
        message_size_ext = self._buffer.get_byte()

        message_size = self._buffer.get_uint()
        message_size += message_size_ext << 32
        return QMessage(None, message_type, message_size, message_compression_mode)


    def read_data(self, message_size, compression_mode = 0, **options):
        '''
        Reads and optionally parses data part of a message.

        .. note:: :func:`.read_header` is required to be called before executing
                  the :func:`.read_data`

        :Parameters:
         - `message_size` (`int`) - size of the message to be read
         - `compression_mode` (`int`) - indicates whether data is compressed, 1 for <2GB, 2 for larger
        :Options:
         - `raw` (`boolean`) - indicates whether read data should parsed or
           returned in raw byte form
         - `numpy_temporals` (`boolean`) - if ``False`` temporal vectors are
           backed by raw q representation (:class:`.QTemporalList`,
           :class:`.QTemporal`) instances, otherwise are represented as
           `numpy datetime64`/`timedelta64` arrays and atoms,
           **Default**: ``False``

        :returns: read data (parsed or raw byte form)
        '''
        self._options = MetaData(**CONVERSION_OPTIONS.union_dict(**options))

        if compression_mode > 0:
            comprHeaderLen = 4 if compression_mode == 1 else 8
            if self._stream:
                self._buffer.wrap(self._read_bytes(comprHeaderLen))
            uncompressed_size = -8 + (self._buffer.get_uint() if compression_mode == 1 else self._buffer.get_long())
            compressed_data = self._read_bytes(message_size - (8+comprHeaderLen)) if self._stream else self._buffer.raw(message_size - (8+comprHeaderLen))

            raw_data = numpy.frombuffer(compressed_data, dtype = numpy.uint8)
            if  uncompressed_size <= 0:
                raise QReaderException('Error while data decompression.')

            raw_data = uncompress(raw_data, numpy.int64(uncompressed_size))
            raw_data = numpy.ndarray.tobytes(raw_data)
            self._buffer.wrap(raw_data)
        elif self._stream:
            raw_data = self._read_bytes(message_size - 8)
            self._buffer.wrap(raw_data)
        if not self._stream and self._options.raw:
            raw_data = self._buffer.raw(message_size - 8)

        if self._options.raw:
            return raw_data
        elif self._options.pandas:
            # Use pandas parsing - choose based on installation type
            from ._pandas import read_object_pandas
            return read_object_pandas(self, raw_data)
        else:
            # Use standard parsing
            return self._read_object()


    def _read_object(self):
        qtype = self._buffer.get_byte()

        reader = self._get_reader(qtype)

        if reader:
            return reader(self, qtype)
        elif qtype >= QBOOL_LIST and qtype <= QTIME_LIST:
            return self._read_list(qtype)
        elif qtype <= QBOOL and qtype >= QTIME:
            return self._read_atom(qtype)

        raise QReaderException('Unable to deserialize q type: %s' % hex(qtype))


    def _get_reader(self, qtype):
        return self._reader_map.get(qtype, None)


    @parse(QERROR)
    def _read_error(self, qtype = QERROR):
        raise QException(self._read_symbol())


    @parse(QSTRING)
    def _read_string(self, qtype = QSTRING):
        self._buffer.skip()  # ignore attributes
        length = self._buffer.get_int()
        raw_bytes = self._buffer.raw(length) if length > 0 else b''
        return String(raw_bytes)


    @parse(QSYMBOL)
    def _read_symbol(self, qtype = QSYMBOL):
        return self._buffer.get_symbol().decode('utf-8')


    @parse(QCHAR)
    def _read_char(self, qtype = QCHAR):
        return Char(chr(self._read_atom(QCHAR)).encode(self._encoding))


    @parse(QGUID)
    def _read_guid(self, qtype = QGUID):
        return uuid.UUID(bytes = self._buffer.raw(16))


    def _read_atom(self, qtype):
        try:
            fmt = STRUCT_MAP[qtype]
            conversion = PY_TYPE[qtype]
            return conversion(self._buffer.get(fmt))
        except KeyError:
            raise QReaderException('Unable to deserialize q type: %s' % hex(qtype))


    @parse(QTIMESPAN, QTIMESTAMP, QTIME, QSECOND, QMINUTE, QDATE, QMONTH, QDATETIME)
    def _read_temporal(self, qtype):
        try:
            fmt = STRUCT_MAP[qtype]
            conversion = PY_TYPE[qtype]
            raw_value = self._buffer.get(fmt)
            return conversion(raw_value)  # conversion now creates the wrapper object directly
        except KeyError:
            raise QReaderException('Unable to deserialize q type: %s' % hex(qtype))


    def _read_list(self, qtype):
        attr = self._buffer.get_byte()
        isLongLength = attr & 0x80 != 0
        length = self._buffer.get_long() if isLongLength else self._buffer.get_uint()
        conversion = PY_TYPE.get(-qtype, None)

        if qtype == QSYMBOL_LIST:
            symbols = self._buffer.get_symbols(length)
            data = numpy.array(symbols, dtype = numpy.bytes_)
            return qlist(data, qtype = qtype, adjust_dtype = False)
        elif qtype == QGUID_LIST:
            data = numpy.array([self._read_guid() for x in range(length)])
            return qlist(data, qtype = qtype, adjust_dtype = False)
        elif qtype in [QBYTE_LIST, QSHORT_LIST, QINT_LIST, QLONG_LIST, 
                       QMONTH_LIST, QDATE_LIST, QDATETIME_LIST, QTIMESPAN_LIST, 
                       QMINUTE_LIST, QSECOND_LIST, QTIME_LIST, QTIMESTAMP_LIST]:
            # For integer and temporal types, manually parse to create wrapper objects
            fmt = STRUCT_MAP[-qtype]  # Get format for individual element
            element_size = ATOM_SIZE[qtype]
            data = []
            for i in range(length):
                raw_element = self._buffer.raw(element_size)
                value = struct.unpack(self._buffer.endianness + fmt, raw_element)[0]
                data.append(conversion(value))
            return qlist(data, qtype = qtype, adjust_dtype = False)
        elif conversion:
            raw = self._buffer.raw(length * ATOM_SIZE[qtype])
            data = numpy.frombuffer(raw, dtype = conversion)
            if not self._is_native:
                data.byteswap(True)


            return qlist(data, qtype = qtype, adjust_dtype = False)
        else:
            raise QReaderException('Unable to deserialize q type: %s' % hex(qtype))


    @parse(QDICTIONARY)
    def _read_dictionary(self, qtype = QDICTIONARY):
        keys = self._read_object()
        values = self._read_object()

        if isinstance(keys, QTable):
            return QKeyedTable(keys, values)
        else:
            return QDictionary(keys, values)


    @parse(QTABLE)
    def _read_table(self, qtype = QTABLE):
        self._buffer.skip()  # ignore attributes
        self._buffer.skip()  # ignore dict type stamp

        columns = self._read_object()
        data = self._read_object()

        return qtable(columns, data, qtype = QTABLE)


    @parse(QGENERAL_LIST)
    def _read_general_list(self, qtype = QGENERAL_LIST):
        self._buffer.skip()  # ignore attributes
        length = self._buffer.get_int()

        return [self._read_object() for x in range(length)]


    @parse(QNULL)
    @parse(QUNARY_FUNC)
    @parse(QBINARY_FUNC)
    @parse(QTERNARY_FUNC)
    def _read_function(self, qtype = QNULL):
        code = self._buffer.get_byte()
        return None if qtype == QNULL and code == 0 else QFunction(qtype)


    @parse(QLAMBDA)
    def _read_lambda(self, qtype = QLAMBDA):
        self._buffer.get_symbol()  # skip
        expression = self._read_object()
        return QLambda(expression.decode())


    @parse(QCOMPOSITION_FUNC)
    def _read_function_composition(self, qtype = QCOMPOSITION_FUNC):
        self._read_projection(qtype)  # skip
        return QFunction(qtype)


    @parse(QADVERB_FUNC_106)
    @parse(QADVERB_FUNC_107)
    @parse(QADVERB_FUNC_108)
    @parse(QADVERB_FUNC_109)
    @parse(QADVERB_FUNC_110)
    @parse(QADVERB_FUNC_111)
    def _read_adverb_function(self, qtype = QADVERB_FUNC_106):
        self._read_object()  # skip
        return QFunction(qtype)


    @parse(QPROJECTION)
    def _read_projection(self, qtype = QPROJECTION):
        length = self._buffer.get_int()
        parameters = [ self._read_object() for x in range(length) ]
        return QProjection(parameters)


    def _read_bytes(self, length):
        if not self._stream:
            raise QReaderException('There is no input data. QReader requires either stream or data chunk')

        if length == 0:
            return b''
        else:
            data = self._stream.read(length)

        if len(data) == 0:
            raise QReaderException('Error while reading data')
        return data



    class BytesBuffer(object):
        '''
        Utility class for reading bytes from wrapped buffer.
        '''

        def __init__(self):
            self._endianness = '@'


        @property
        def endianness(self):
            '''
            Gets the endianness.
            '''
            return self._endianness


        @endianness.setter
        def endianness(self, endianness):
            '''
            Sets the byte order (endianness) for reading from the buffer.

            :Parameters:
             - `endianness` (``<`` or ``>``) - byte order indicator
            '''
            self._endianness = endianness


        def wrap(self, data):
            '''
            Wraps the data in the buffer.

            :Parameters:
             - `data` - data to be wrapped
            '''
            self._data = data
            self._position = 0
            self._size = len(data)


        def skip(self, offset = 1):
            '''
            Skips reading of `offset` bytes.

            :Parameters:
             - `offset` (`integer`) - number of bytes to be skipped
            '''
            new_position = self._position + offset

            if new_position > self._size:
                raise QReaderException('Attempt to read data out of buffer bounds')

            self._position = new_position


        def raw(self, offset):
            '''
            Gets `offset` number of raw bytes.

            :Parameters:
             - `offset` (`integer`) - number of bytes to be retrieved

            :returns: raw bytes
            '''
            new_position = self._position + offset

            if new_position > self._size:
                raise QReaderException('Attempt to read data out of buffer bounds')

            raw = self._data[self._position : new_position]
            self._position = new_position
            return raw


        def get(self, fmt, offset = None):
            '''
            Gets bytes from the buffer according to specified format or `offset`.

            :Parameters:
             - `fmt` (struct format) - conversion to be applied for reading
             - `offset` (`integer`) - number of bytes to be retrieved

            :returns: unpacked bytes
            '''
            fmt = self.endianness + fmt
            offset = offset if offset else struct.calcsize(fmt)
            return struct.unpack(fmt, self.raw(offset))[0]


        def get_byte(self):
            '''
            Gets a single byte from the buffer.

            :returns: single byte
            '''
            return self.get('b')


        def get_int(self):
            '''
            Gets a single 32-bit integer from the buffer.

            :returns: single integer
            '''
            return self.get('i')


        def get_uint(self):
            '''
            Gets a single 32-bit unsigned integer from the buffer.

            :returns: single integer
            '''
            return self.get('I')


        def get_long(self):
            '''
            Gets a single 64-bit integer from the buffer.

            :returns: single integer
            '''
            return self.get('q')


        def get_symbol(self):
            '''
            Gets a single, ``\\x00`` terminated string from the buffer.

            :returns: ``\\x00`` terminated string
            '''
            new_position = self._data.find(b'\x00', self._position)

            if new_position < 0:
                raise QReaderException('Failed to read symbol from stream')

            raw = self._data[self._position : new_position]
            self._position = new_position + 1
            return raw


        def get_symbols(self, count):
            '''
            Gets ``count`` ``\\x00`` terminated strings from the buffer.

            :Parameters:
             - `count` (`integer`) - number of strings to be read

            :returns: list of ``\\x00`` terminated string read from the buffer
            '''
            c = 0
            new_position = self._position

            if count == 0:
                return []

            while c < count:
                new_position = self._data.find(b'\x00', new_position)

                if new_position < 0:
                    raise QReaderException('Failed to read symbol from stream')

                c += 1
                new_position += 1

            raw = self._data[self._position : new_position - 1]
            self._position = new_position

            return raw.split(b'\x00')





#
#  Copyright (c) 2011-2014 Exxeleron GmbH
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import struct
try:
    from cStringIO import BytesIO
except ImportError:
    from io import BytesIO

from .qmetaconversion import MetaData, CONVERSION_OPTIONS
from .qtype import *  # @UnusedWildImport
from .qcollection import qlist, QList, QTemporalList, QDictionary, QTable, QKeyedTable, get_list_qtype
from .qtemporal import QTemporal, to_raw_qtemporal, array_to_raw_qtemporal


class QWriterException(Exception):
    '''
    Indicates an error raised during data serialization.
    '''
    pass



ENDIANNESS = '\1' if sys.byteorder == 'little' else '\0'


class QWriter(object):
    '''
    Provides serialization to q IPC protocol.
    
    :Parameters:
     - `stream` (`socket` or `None`) - stream for data serialization
     - `protocol_version` (`integer`) - version IPC protocol
     - `encoding` (`string`) - encoding for characters serialization
    
    :Attrbutes:
     - `_writer_map` - stores mapping between Python types and functions 
       responsible for serializing into IPC representation
    '''

    _writer_map = {}
    serialize = Mapper(_writer_map)


    def __init__(self, stream, protocol_version, encoding = 'latin-1'):
        self._stream = stream
        self._protocol_version = protocol_version
        self._encoding = encoding


    def write(self, data, msg_type, **options):
        '''Serializes and pushes single data object to a wrapped stream.
        
        :Parameters:
         - `data` - data to be serialized
         - `msg_type` (one of the constants defined in :class:`.MessageType`) -
           type of the message
        :Options:
         - `single_char_strings` (`boolean`) - if ``True`` single char Python 
           strings are encoded as q strings instead of chars, 
           **Default**: ``False``
        
        :returns: if wraped stream is ``None`` serialized data, 
                  otherwise ``None`` 
        '''
        self._buffer = BytesIO()

        self._options = MetaData(**CONVERSION_OPTIONS.union_dict(**options))

        # header and placeholder for message size
        self._buffer.write(('%s%s\0\0\0\0\0\0' % (ENDIANNESS, chr(msg_type))).encode(self._encoding))

        self._write(data)

        # update message size
        data_size = self._buffer.tell()
        self._buffer.seek(4)
        self._buffer.write(struct.pack('i', data_size))

        # write data to socket
        if self._stream:
            self._stream.sendall(self._buffer.getvalue())
        else:
            return self._buffer.getvalue()


    def _write(self, data):
        if data is None:
            self._write_null()
        else:
            if isinstance(data, Exception) or (type(data) == type and issubclass(data, Exception)):
                data_type = Exception
            else:
                data_type = type(data)

            writer = self._get_writer(data_type)

            if writer:
                writer(self, data)
            else:
                qtype = Q_TYPE.get(type(data), None)

                if qtype:
                    self._write_atom(data, qtype)
                else:
                    raise QWriterException('Unable to serialize type: %s' % data.__class__ if isinstance(data, object) else type(data))


    def _get_writer(self, data_type):
        return self._writer_map.get(data_type, None)


    def _write_null(self):
        self._buffer.write(struct.pack('=bx', QNULL))


    @serialize(Exception)
    def _write_error(self, data):
        self._buffer.write(struct.pack('b', QERROR))
        if isinstance(data, Exception):
            msg = data.__class__.__name__
            if data.args:
                msg = data.args[0]
        else:
            msg = data.__name__

        self._buffer.write(msg.encode(self._encoding))
        self._buffer.write(b'\0')


    def _write_atom(self, data, qtype):
        try:
            self._buffer.write(struct.pack('b', qtype))
            fmt = STRUCT_MAP[qtype]
            if type(data) == numpy.bool_:
                self._buffer.write(struct.pack(fmt, int(data)))
            else:
                self._buffer.write(struct.pack(fmt, data))
        except KeyError:
            raise QWriterException('Unable to serialize type: %s' % data.__class__ if isinstance(data, object) else type(data))


    @serialize(tuple, list)
    def _write_generic_list(self, data):
        self._buffer.write(struct.pack('=bxi', QGENERAL_LIST, len(data)))
        for element in data:
            self._write(element)


    @serialize(str, bytes)
    def _write_string(self, data):
        if not self._options.single_char_strings and len(data) == 1:
            self._write_atom(ord(data), QCHAR)
        else:
            self._buffer.write(struct.pack('=bxi', QSTRING, len(data)))
            if isinstance(data, str):
                self._buffer.write(data.encode(self._encoding))
            else:
                self._buffer.write(data)


    @serialize(numpy.bytes_)
    def _write_symbol(self, data):
        self._buffer.write(struct.pack('=b', QSYMBOL))
        if data:
            self._buffer.write(data)
        self._buffer.write(b'\0')


    @serialize(uuid.UUID)
    def _write_guid(self, data):
        if self._protocol_version < 3:
            raise QWriterException('kdb+ protocol version violation: Guid not supported pre kdb+ v3.0')

        self._buffer.write(struct.pack('=b', QGUID))
        self._buffer.write(data.bytes)


    @serialize(QTemporal)
    def _write_temporal(self, data):
        try:
            if self._protocol_version < 1 and (data.meta.qtype == QTIMESPAN or data.meta.qtype == QTIMESTAMP):
                raise QWriterException('kdb+ protocol version violation: data type %s not supported pre kdb+ v2.6' % hex(data.meta.qtype))

            self._buffer.write(struct.pack('=b', data.meta.qtype))
            fmt = STRUCT_MAP[data.meta.qtype]
            self._buffer.write(struct.pack(fmt, to_raw_qtemporal(data.raw, data.meta.qtype)))
        except KeyError:
            raise QWriterException('Unable to serialize type: %s' % type(data))


    @serialize(numpy.datetime64, numpy.timedelta64)
    def _write_numpy_temporal(self, data):
        try:
            qtype = TEMPORAL_PY_TYPE[str(data.dtype)]

            if self._protocol_version < 1 and (qtype == QTIMESPAN or qtype == QTIMESTAMP):
                raise QWriterException('kdb+ protocol version violation: data type %s not supported pre kdb+ v2.6' % hex(qtype))

            self._buffer.write(struct.pack('=b', qtype))
            fmt = STRUCT_MAP[qtype]
            self._buffer.write(struct.pack(fmt, to_raw_qtemporal(data, qtype)))
        except KeyError:
            raise QWriterException('Unable to serialize type: %s' % data.dtype)


    @serialize(QLambda)
    def _write_lambda(self, data):
        self._buffer.write(struct.pack('=b', QLAMBDA))
        self._buffer.write(b'\0')
        self._write_string(data.expression)


    @serialize(QProjection)
    def _write_projection(self, data):
        self._buffer.write(struct.pack('=bi', QPROJECTION, len(data.parameters)))
        for parameter in data.parameters:
            self._write(parameter)


    @serialize(QDictionary, QKeyedTable)
    def _write_dictionary(self, data):
        self._buffer.write(struct.pack('=b', QDICTIONARY))
        self._write(data.keys)
        self._write(data.values)


    @serialize(QTable)
    def _write_table(self, data):
        self._buffer.write(struct.pack('=bxb', QTABLE, QDICTIONARY))
        self._write(qlist(numpy.array(data.dtype.names), qtype = QSYMBOL_LIST))
        self._buffer.write(struct.pack('=bxi', QGENERAL_LIST, len(data.dtype)))
        for column in data.dtype.names:
            self._write_list(data[column], data.meta[column])


    @serialize(numpy.ndarray, QList, QTemporalList)
    def _write_list(self, data, qtype = None):
        if qtype is not None:
            qtype = -abs(qtype)

        if qtype is None:
            qtype = get_list_qtype(data)

        if self._protocol_version < 1 and (abs(qtype) == QTIMESPAN_LIST or abs(qtype) == QTIMESTAMP_LIST):
            raise QWriterException('kdb+ protocol version violation: data type %s not supported pre kdb+ v2.6' % hex(data.meta.qtype))

        if qtype == QGENERAL_LIST:
            self._write_generic_list(data)
        elif qtype == QCHAR:
            self._write_string(data.tobytes())
        else:
            self._buffer.write(struct.pack('=bxi', -qtype, len(data)))
            if data.dtype.type in (numpy.datetime64, numpy.timedelta64):
                # convert numpy temporal to raw q temporal
                data = array_to_raw_qtemporal(data, qtype = qtype)

            if qtype == QSYMBOL:
                for symbol in data:
                    if symbol:
                        self._buffer.write(symbol)
                    self._buffer.write(b'\0')
            elif qtype == QGUID:
                if self._protocol_version < 3:
                    raise QWriterException('kdb+ protocol version violation: Guid not supported pre kdb+ v3.0')

                for guid in data:
                    self._buffer.write(guid.bytes)
            else:
                self._buffer.write(data.tobytes())




"""
Configuration management for qmcp
"""
import os
import shutil
import sys
from pathlib import Path
from platformdirs import user_config_dir
import datetime

# Handle tomllib for different Python versions
if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

# Global config cache
_config = None
_config_creation_failed = False
_config_creation_error = None
_is_full_installation = None
_output_format_fallback_warning = None

# Debug logging control
DEBUG = False

def _log_debug(message):
    """Log debug messages to file for tracking config loading"""
    if not DEBUG:
        return
    try:
        with open('/tmp/qmcp_config_debug.log', 'a') as f:
            timestamp = datetime.datetime.now().strftime('%H:%M:%S.%f')
            f.write(f'[{timestamp}] {message}\n')
            f.flush()
    except:
        pass  # Ignore logging errors

def detect_full_installation():
    """Detect if this is a full installation (has qython and enhanced qpython)"""
    global _is_full_installation
    _log_debug(f"detect_full_installation() called, cached: {_is_full_installation}")
    if _is_full_installation is None:
        try:
            # Test for environment variable simulation
            import os
            if os.getenv('QMCP_SIMULATE_THIN'):
                raise ImportError("Simulating thin package")

            # Test for the key missing piece in thin builds - qpython.display
            from .qpython import display
            _is_full_installation = True
        except ImportError:
            _is_full_installation = False

    _log_debug(f"detect_full_installation() result: {_is_full_installation}")
    return _is_full_installation

def get_config_path():
    """Get the path to the config file, with environment override support"""
    # Allow override via environment variable
    env_path = os.getenv('QMCP_CONFIG_PATH')
    if env_path:
        return Path(env_path)

    # Default to environment-scoped config for venv or conda
    env_path = os.getenv('VIRTUAL_ENV') or os.getenv('CONDA_PREFIX')
    if env_path:
        return Path(env_path) / '.qmcp' / 'config.toml'

    # Fallback: system config directory
    config_dir = Path(user_config_dir("qmcp"))
    return config_dir / "config.toml"

def ensure_config():
    """Ensure config file exists, create from template if needed"""
    global _config_creation_failed, _config_creation_error
    config_path = get_config_path()

    if not config_path.exists():
        try:
            # Create config directory
            config_path.parent.mkdir(parents=True, exist_ok=True)

            # Copy template (content only, no permissions)
            template_path = Path(__file__).parent / "config_template.toml"
            shutil.copyfile(template_path, config_path)

            print(f"✅ Created qmcp config at: {config_path}")
            print("💡 Edit this file to customize your setup")
        except Exception as e:
            _config_creation_failed = True
            _config_creation_error = f"Failed to create config at {config_path}: {e}"
            print(f"❌ {_config_creation_error}")

    return config_path

def load_config():
    """Load configuration from file with automatic fallbacks"""
    global _output_format_fallback_warning
    _log_debug("load_config() called")
    _output_format_fallback_warning = None  # Clear any previous warning
    config_path = ensure_config()

    try:
        with open(config_path, 'rb') as f:
            config = tomllib.load(f)
    except Exception as e:
        print(f"❌ Error loading config from {config_path}: {e}")
        print("💡 Using default configuration")
        config = get_default_config()

    # Check for incompatible qython + thin + no-cloud combination
    output_format = config.get("default", {}).get("output_format", "qython")
    cloud_enabled = config.get("default", {}).get("cloud_enabled", False)
    full_installation = detect_full_installation()

    _log_debug(f"Config check: output_format={output_format}, cloud_enabled={cloud_enabled}, full_installation={full_installation}")

    if output_format == "qython" and not full_installation and not cloud_enabled:
        _output_format_fallback_warning = "⚠️  Qython formatting requires cloud services or full installation. Switched to raw q output. Enable cloud_enabled=true or install full version for enhanced formatting."
        # Override the config to use q format
        if "default" not in config:
            config["default"] = {}
        config["default"]["output_format"] = "q"
        _log_debug(f"Applied fallback: warning set, config overridden to q format")
    else:
        _log_debug(f"No fallback needed")

    _log_debug(f"load_config() returning, warning={_output_format_fallback_warning is not None}")
    return config

def get_default_config():
    """Get default configuration when config file is unavailable"""
    return {
        "default": {
            "output_format": "qython",
            "cloud_enabled": False
        },
        "servers": {
            "default": {
                "host": "localhost",
                "port": 5001
            }
        }
    }

def get_config():
    """Get global configuration (cached)"""
    global _config
    _log_debug(f"get_config() called, cached: {_config is not None}")
    if _config is None:
        _config = load_config()
    return _config

def is_cloud_enabled():
    """Check if cloud translation is enabled"""
    config = get_config()
    return config.get("default", {}).get("cloud_enabled", False)


def get_config_creation_error():
    """Get config creation error message if it failed"""
    return _config_creation_error if _config_creation_failed else None

def get_output_format_fallback_warning():
    """Get output format fallback warning if it occurred"""
    _log_debug(f"get_output_format_fallback_warning() called, warning: {_output_format_fallback_warning is not None}")
    return _output_format_fallback_warning

def get_output_format():
    """Get output format setting"""
    config = get_config()
    return config.get("default", {}).get("output_format", "qython")

def get_server_config(server_name="default"):
    """Get configuration for a specific server"""
    config = get_config()
    servers = config.get("servers", {})
    return servers.get(server_name, servers.get("default", {
        "host": "localhost",
        "port": 5001
    }))

def get_operational_defaults():
    """
    Get operational defaults from [default] section

    These settings can be overridden per-server but provide global defaults
    for timeouts, print behavior, and console settings.

    Returns:
        Dict with operational settings
    """
    config = get_config()
    defaults = config.get("default", {})

    return {
        "connection_timeout": defaults.get("connection_timeout", 2),
        "async_timeout": defaults.get("async_timeout", 1),
        "interrupt_timeout": defaults.get("interrupt_timeout", 10),
        "print_to_async": defaults.get("print_to_async", True),
        "console_size": defaults.get("console_size", []),  # Empty array = don't change
    }

def get_server_config_with_defaults(server_name="default"):
    """
    Get server configuration merged with operational defaults

    Args:
        server_name: Name of server configuration to load

    Returns:
        Dict with merged config (operational defaults + server specifics)
        Server-specific settings override operational defaults
    """
    # Get operational defaults
    operational_defaults = get_operational_defaults()

    # Get server-specific config
    server_config = get_server_config(server_name)

    # Merge: operational defaults + server config
    # Server config overrides defaults
    return {**operational_defaults, **server_config}

def build_connection_string(server_config):
    """
    Build q connection string from server config

    Args:
        server_config: Dict with host, port, optional user/password

    Returns:
        Connection string in format: host:port or host:port:user:password
    """
    host = server_config.get("host", "localhost")
    port = server_config.get("port", 5001)
    user = server_config.get("user", "")
    password = server_config.get("password", "")

    # Basic connection string
    conn_str = f"{host}:{port}"

    # Add authentication if provided
    if user and password:
        conn_str = f"{conn_str}:{user}:{password}"

    return conn_str


#!/usr/bin/env python3
"""
Authentication tools for qmcp cloud services
"""
import requests
from .config import get_config

# Cloud server URL
CLOUD_SERVER_URL = "https://qython.dev"


def request_api_key(email: str) -> str:
    """
    Request an API key for cloud translation services.

    An API key enables cloud-based Qython translation (Python-like syntax → q code).
    The API key will be emailed to you.

    Args:
        email: Your email address

    Returns:
        Status message indicating success or instructions for next steps

    Example:
        >>> request_api_key("user@example.com")
        "✅ API key sent to user@example.com. Check your inbox (and spam folder)."
    """
    try:
        # Call the cloud server endpoint
        response = requests.post(
            f"{CLOUD_SERVER_URL}/api/request-key",
            json={"email": email},
            timeout=30
        )

        if response.status_code == 200:
            data = response.json()
            if data.get("success"):
                return f"✅ {data.get('message')}\n\nOnce you receive the email, add the API key to your config by telling your AI assistant or manually editing the config file."
            else:
                return f"❌ {data.get('message', 'Unknown error')}"
        else:
            # Try to parse error message
            try:
                error_data = response.json()
                error_msg = error_data.get("detail", response.text)
            except:
                error_msg = response.text

            return f"❌ Request failed (HTTP {response.status_code}): {error_msg}"

    except requests.exceptions.Timeout:
        return "❌ Request timed out. Please check your internet connection and try again."
    except requests.exceptions.ConnectionError:
        return f"❌ Could not connect to {CLOUD_SERVER_URL}. Please check your internet connection."
    except Exception as e:
        return f"❌ Unexpected error: {str(e)}"



#
#  Copyright (c) 2011-2014 Exxeleron GmbH
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from .qmetaconversion import MetaData
from .qtype import *  # @UnusedWildImport
from numpy import longlong

_MILLIS_PER_DAY = 24 * 60 * 60 * 1000
_MILLIS_PER_DAY_FLOAT = float(_MILLIS_PER_DAY)
_QEPOCH_MS = long(10957 * _MILLIS_PER_DAY)
_EPOCH_QTIMESTAMP_NS = _QEPOCH_MS * 1000000


_EPOCH_QMONTH = numpy.datetime64('2000-01', 'M')
_EPOCH_QDATE = numpy.datetime64('2000-01-01', 'D')
_EPOCH_QDATETIME = numpy.datetime64(_QEPOCH_MS, 'ms')
_EPOCH_TIMESTAMP = numpy.datetime64(_EPOCH_QTIMESTAMP_NS, 'ns')


_QMONTH_NULL = qnull(QMONTH)
_QDATE_NULL = qnull(QDATE)
_QDATETIME_NULL = qnull(QDATETIME)
_QMINUTE_NULL = qnull(QMINUTE)
_QSECOND_NULL = qnull(QSECOND)
_QTIME_NULL = qnull(QTIME)
_QTIMESTAMP_NULL = qnull(QTIMESTAMP)
_QTIMESPAN_NULL = qnull(QTIMESPAN)



class QTemporal(object):
    '''
    Represents a q temporal value.
    
    The :class:`.QTemporal` wraps `numpy.datetime64` or `numpy.timedelta64`
    along with meta-information like qtype indicator.
    
    :Parameters:
     - `dt` (`numpy.datetime64` or `numpy.timedelta64`) - datetime to be wrapped
    '''

    def __init__(self, dt):
        self._datetime = dt

    def _meta_init(self, **meta):
        self.meta = MetaData(**meta)

    @property
    def raw(self):
        '''Return wrapped datetime object.
        
        :returns: `numpy.datetime64` or `numpy.timedelta64` - wrapped datetime
        '''
        return self._datetime

    def __str__(self):
        return '%s' % (self._datetime)

    def __repr__(self):
        return '%s [%s]' % (self._datetime, self.meta)

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
            and self.meta.qtype == other.meta.qtype
            and self._datetime == other._datetime)

    def __ne__(self, other):
        return not self.__eq__(other)



def qtemporal(dt, **meta):
    '''Converts a `numpy.datetime64` or `numpy.timedelta64` to 
    :class:`.QTemporal` and enriches object instance with given meta data.
    
    Examples:
    
       >>> qtemporal(numpy.datetime64('2001-01-01', 'D'), qtype=QDATE)
       2001-01-01 [metadata(qtype=-14)]
       >>> qtemporal(numpy.timedelta64(43499123, 'ms'), qtype=QTIME)
       43499123 milliseconds [metadata(qtype=-19)]
       >>> qtemporal(qnull(QDATETIME), qtype=QDATETIME)
       nan [metadata(qtype=-15)]
    
    :Parameters:
     - `dt` (`numpy.datetime64` or `numpy.timedelta64`) - datetime to be wrapped
    :Kwargs:
     - `qtype` (`integer`) - qtype indicator
    
    :returns: `QTemporal` - wrapped datetime 
    '''
    result = QTemporal(dt)
    result._meta_init(**meta)
    return result



def from_raw_qtemporal(raw, qtype):
    '''
    Converts raw numeric value to `numpy.datetime64` or `numpy.timedelta64`
    instance.
    
    Actual conversion applied to raw numeric value depends on `qtype` parameter.
    
    :Parameters:
     - `raw` (`integer`, `float`) - raw representation to be converted
     - `qtype` (`integer`) - qtype indicator
     
    :returns: `numpy.datetime64` or `numpy.timedelta64` - converted datetime
    '''
    return _FROM_Q[qtype](raw)



def to_raw_qtemporal(dt, qtype):
    '''
    Converts datetime/timedelta instance to raw numeric value.
    
    Actual conversion applied to datetime/timedelta instance depends on `qtype` 
    parameter.
    
    :Parameters:
     - `dt` (`numpy.datetime64` or `numpy.timedelta64`) - datetime/timedelta
       object to be converted
     - `qtype` (`integer`) - qtype indicator
     
    :returns: `integer`, `float` - raw numeric value
    '''
    return _TO_Q[qtype](dt)



def array_from_raw_qtemporal(raw, qtype):
    '''
    Converts `numpy.array` containing raw q representation to ``datetime64``/``timedelta64``
    array.
    
    Examples:
    
      >>> raw = numpy.array([366, 121, qnull(QDATE)])
      >>> print(array_from_raw_qtemporal(raw, qtype = QDATE))
      ['2001-01-01' '2000-05-01' 'NaT']
    
    :Parameters:
     - `raw` (`numpy.array`) - numpy raw array to be converted
     - `qtype` (`integer`) - qtype indicator
    
    :returns: `numpy.array` - numpy array with ``datetime64``/``timedelta64``
    
    :raises: `ValueError`
    '''
    if not isinstance(raw, numpy.ndarray):
        raise ValueError('raw parameter is expected to be of type: numpy.ndarray. Was: %s' % type(raw))

    qtype = -abs(qtype)
    conversion = _FROM_RAW_LIST[qtype]

    mask = raw == qnull(qtype)

    dtype = PY_TYPE[qtype]
    array = raw.astype(dtype) if dtype != raw.dtype else raw

    array = conversion(array) if conversion else array
    null = _NUMPY_NULL[qtype]
    array = numpy.where(mask, null, array)
    return array



def array_to_raw_qtemporal(array, qtype):
    '''
    Converts `numpy.array` containing ``datetime64``/``timedelta64`` to raw
    q representation.
    
    Examples:
    
      >>> na_dt = numpy.arange('1999-01-01', '2005-12-31', dtype='datetime64[D]')
      >>> print(array_to_raw_qtemporal(na_dt, qtype = QDATE_LIST))
      [-365 -364 -363 ..., 2188 2189 2190]
      >>> array_to_raw_qtemporal(numpy.arange(-20, 30, dtype='int32'), qtype = QDATE_LIST)
      Traceback (most recent call last):
        ...
      ValueError: array.dtype is expected to be of type: datetime64 or timedelta64. Was: int32
    
    :Parameters:
     - `array` (`numpy.array`) - numpy datetime/timedelta array to be converted
     - `qtype` (`integer`) - qtype indicator
    
    :returns: `numpy.array` - numpy array with raw values
    
    :raises: `ValueError`
    '''
    if not isinstance(array, numpy.ndarray):
        raise ValueError('array parameter is expected to be of type: numpy.ndarray. Was: %s' % type(array))

    if not array.dtype.type in (numpy.datetime64, numpy.timedelta64):
        raise ValueError('array.dtype is expected to be of type: datetime64 or timedelta64. Was: %s' % array.dtype)

    qtype = -abs(qtype)
    conversion = _TO_RAW_LIST[qtype]
    raw = array.view(numpy.int64).view(numpy.ndarray)
    mask = raw == numpy.int64(-2 ** 63)

    raw = conversion(raw) if conversion else raw
    null = qnull(qtype)
    raw = numpy.where(mask, null, raw)
    return raw



def _from_qmonth(raw):
    if raw == _QMONTH_NULL:
        return _NUMPY_NULL[QMONTH]
    else:
        return _EPOCH_QMONTH + numpy.timedelta64(int(raw), 'M')



def _to_qmonth(dt):
    t_dt = type(dt)
    if t_dt == numpy.int32:
        return dt
    elif t_dt == numpy.datetime64:
        return (dt - _EPOCH_QMONTH).astype(int) if not numpy.isnan(dt) else _QMONTH_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



def _from_qdate(raw):
    if raw == _QDATE_NULL:
        return _NUMPY_NULL[QDATE]
    else:
        return _EPOCH_QDATE + numpy.timedelta64(int(raw), 'D')



def _to_qdate(dt):
    t_dt = type(dt)
    if t_dt == numpy.int32:
        return dt
    elif t_dt == numpy.datetime64:
        return (dt - _EPOCH_QDATE).astype(int) if not numpy.isnan(dt) else _QDATE_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



def _from_qdatetime(raw):
    if numpy.isnan(raw) or raw == _QDATETIME_NULL:
        return _NUMPY_NULL[QDATETIME]
    else:
        return _EPOCH_QDATETIME + numpy.timedelta64(long(_MILLIS_PER_DAY * raw), 'ms')



def _to_qdatetime(dt):
    t_dt = type(dt)
    if t_dt == numpy.float64:
        return dt
    elif t_dt == numpy.datetime64:
        return (dt - _EPOCH_QDATETIME).astype(float) / _MILLIS_PER_DAY if not numpy.isnan(dt) else _QDATETIME_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



def _from_qminute(raw):
    if raw == _QMINUTE_NULL:
        return _NUMPY_NULL[QMINUTE]
    else:
        return numpy.timedelta64(int(raw), 'm')



def _to_qminute(dt):
    t_dt = type(dt)
    if t_dt == numpy.int32:
        return dt
    elif t_dt == numpy.timedelta64:
        return dt.astype(int) if not numpy.isnan(dt) else _QMINUTE_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



def _from_qsecond(raw):
    if raw == _QSECOND_NULL:
        return _NUMPY_NULL[QSECOND]
    else:
        return numpy.timedelta64(int(raw), 's')



def _to_qsecond(dt):
    t_dt = type(dt)
    if t_dt == numpy.int32:
        return dt
    elif t_dt == numpy.timedelta64:
        return dt.astype(int) if not numpy.isnan(dt) else _QSECOND_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



def _from_qtime(raw):
    if raw == _QTIME_NULL:
        return _NUMPY_NULL[QTIME]
    else:
        return numpy.timedelta64(int(raw), 'ms')



def _to_qtime(dt):
    t_dt = type(dt)
    if t_dt == numpy.int32:
        return dt
    elif t_dt == numpy.timedelta64:
        return dt.astype(int) if not numpy.isnan(dt) else _QTIME_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



def _from_qtimestamp(raw):
    if raw == _QTIMESTAMP_NULL:
        return _NUMPY_NULL[QTIMESTAMP]
    else:
        return _EPOCH_TIMESTAMP + numpy.timedelta64(long(raw), 'ns')



def _to_qtimestamp(dt):
    t_dt = type(dt)
    if t_dt == numpy.int64:
        return dt
    elif t_dt == numpy.datetime64:
        return (dt - _EPOCH_TIMESTAMP).astype(longlong) if not numpy.isnan(dt) else _QTIMESTAMP_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



def _from_qtimespan(raw):
    if raw == _QTIMESPAN_NULL:
        return _NUMPY_NULL[QTIMESPAN]
    else:
        return numpy.timedelta64(long(raw), 'ns')



def _to_qtimespan(dt):
    t_dt = type(dt)
    if t_dt == numpy.int64:
        return dt
    elif t_dt == numpy.timedelta64:
        return dt.astype(longlong) if not numpy.isnan(dt) else _QTIMESTAMP_NULL
    else:
        raise ValueError('Cannot convert %s of type %s to q value.' % (dt, type(dt)))



_FROM_Q = {
           QMONTH:          _from_qmonth,
           QDATE:           _from_qdate,
           QDATETIME:       _from_qdatetime,
           QMINUTE:         _from_qminute,
           QSECOND:         _from_qsecond,
           QTIME:           _from_qtime,
           QTIMESTAMP:      _from_qtimestamp,
           QTIMESPAN:       _from_qtimespan,
           }



_TO_Q = {
         QMONTH:            _to_qmonth,
         QDATE:             _to_qdate,
         QDATETIME:         _to_qdatetime,
         QMINUTE:           _to_qminute,
         QSECOND:           _to_qsecond,
         QTIME:             _to_qtime,
         QTIMESTAMP:        _to_qtimestamp,
         QTIMESPAN:         _to_qtimespan,
         }


_TO_RAW_LIST = {
                QMONTH:      lambda a: (a - 360).astype(numpy.int32),
                QDATE:       lambda a: (a - 10957).astype(numpy.int32),
                QDATETIME:   lambda a: ((a - _QEPOCH_MS) / _MILLIS_PER_DAY_FLOAT).astype(numpy.float64),
                QMINUTE:     lambda a: a.astype(numpy.int32),
                QSECOND:     lambda a: a.astype(numpy.int32),
                QTIME:       lambda a: a.astype(numpy.int32),
                QTIMESTAMP:  lambda a: a - _EPOCH_QTIMESTAMP_NS,
                QTIMESPAN:   None,
                }



_FROM_RAW_LIST = {
                  QMONTH:      lambda a: numpy.array((a + 360), dtype = 'datetime64[M]'),
                  QDATE:       lambda a: numpy.array((a + 10957), dtype = 'datetime64[D]'),
                  QDATETIME:   lambda a: numpy.array((a * _MILLIS_PER_DAY + _QEPOCH_MS), dtype = 'datetime64[ms]'),
                  QMINUTE:     lambda a: numpy.array(a, dtype = 'timedelta64[m]'),
                  QSECOND:     lambda a: numpy.array(a, dtype = 'timedelta64[s]'),
                  QTIME:       lambda a: numpy.array(a, dtype = 'timedelta64[ms]'),
                  QTIMESTAMP:  lambda a: numpy.array((a + _EPOCH_QTIMESTAMP_NS), dtype = 'datetime64[ns]'),
                  QTIMESPAN:   lambda a: numpy.array(a, dtype = 'timedelta64[ns]'),
                  }



_NUMPY_NULL = {
               QMONTH:      numpy.datetime64('NaT', 'M'),
               QDATE:       numpy.datetime64('NaT', 'D'),
               QDATETIME:   numpy.datetime64('NaT', 'ms'),
               QMINUTE:     numpy.timedelta64('NaT', 'm'),
               QSECOND:     numpy.timedelta64('NaT', 's'),
               QTIME:       numpy.timedelta64('NaT', 'ms'),
               QTIMESTAMP:  numpy.datetime64('NaT', 'ns'),
               QTIMESPAN:   numpy.timedelta64('NaT', 'ns'),
               }



#
#  Copyright (c) 2011-2014 Exxeleron GmbH
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from .qtype import *  # @UnusedWildImport
from .qmetaconversion import MetaData
from .qtemporal import qtemporal, QTemporal, from_raw_qtemporal, to_raw_qtemporal


class QList(numpy.ndarray):
    '''An array object represents a q vector.'''

    def _meta_init(self, **meta):
        '''Initialises the meta-information.'''
        self.meta = MetaData(**meta)
    
    @property 
    def dtype(self):
        """Return custom dtype for integer and temporal QList arrays to display nicely in pandas"""
        # Import here to avoid circular imports
        from .qtypes import (QInt8Dtype, QInt16Dtype, QInt32Dtype, QInt64Dtype,
                           QDateDayDtype, QDateMonthDtype, QTimeMinuteDtype, QTimeSecondDtype,
                           QTimeMsDtype, QTimeNsDtype, QDatetimeMsDtype, QDatetimeNsDtype)
        from .qtype import (QBYTE_LIST, QSHORT_LIST, QINT_LIST, QLONG_LIST,
                          QMONTH_LIST, QDATE_LIST, QDATETIME_LIST, QTIMESPAN_LIST,
                          QMINUTE_LIST, QSECOND_LIST, QTIME_LIST, QTIMESTAMP_LIST)
        
        if hasattr(self, 'meta') and hasattr(self.meta, 'qtype'):
            qtype = abs(self.meta.qtype)  # Use absolute value to handle negative qtypes
            # Integer types
            if qtype == QBYTE_LIST:
                return QInt8Dtype()
            elif qtype == QSHORT_LIST:
                return QInt16Dtype()
            elif qtype == QINT_LIST:
                return QInt32Dtype()
            elif qtype == QLONG_LIST:
                return QInt64Dtype()
            # Temporal types
            elif qtype == QMONTH_LIST:
                return QDateMonthDtype()
            elif qtype == QDATE_LIST:
                return QDateDayDtype()
            elif qtype == QDATETIME_LIST:
                return QDatetimeMsDtype()
            elif qtype == QTIMESPAN_LIST:
                return QTimeNsDtype()
            elif qtype == QMINUTE_LIST:
                return QTimeMinuteDtype()
            elif qtype == QSECOND_LIST:
                return QTimeSecondDtype()
            elif qtype == QTIME_LIST:
                return QTimeMsDtype()
            elif qtype == QTIMESTAMP_LIST:
                return QDatetimeNsDtype()
        
        # For other types, use original numpy dtype
        return super().dtype
    
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Handle numpy universal functions on QList arrays for pandas compatibility."""
        import numpy
        
        # Check if this QList contains QInteger or QTemporal objects
        has_wrapper_objects = (hasattr(self, 'meta') and hasattr(self.meta, 'qtype') and 
                              abs(self.meta.qtype) in [4, 5, 6, 7, 12, 13, 14, 15, 16, 17, 18, 19])  
                              # QBYTE_LIST, QSHORT_LIST, QINT_LIST, QLONG_LIST, 
                              # QTIMESTAMP_LIST, QMONTH_LIST, QDATE_LIST, QDATETIME_LIST, 
                              # QTIMESPAN_LIST, QMINUTE_LIST, QSECOND_LIST, QTIME_LIST
        
        if has_wrapper_objects and method == '__call__':
            if ufunc is numpy.isnan:
                # Return array of boolean values for each element
                return numpy.array([elem.is_null() if hasattr(elem, 'is_null') else False for elem in self])
            elif ufunc is numpy.isinf:
                # Return array of boolean values for infinity check  
                return numpy.array([elem.is_inf() if hasattr(elem, 'is_inf') else False for elem in self])
            elif ufunc is numpy.isfinite:
                # Return array of boolean values for finite check
                return numpy.array([elem.is_finite() if hasattr(elem, 'is_finite') else True for elem in self])
        
        # For other cases, delegate to parent
        return super().__array_ufunc__(ufunc, method, *inputs, **kwargs)

    def __eq__(self, other):
        if (type(other) == QList or
            type(other).__module__ == 'pandas.core.series' and type(other).__name__ == 'Series'):
                return numpy.array_equal(self, other)
        return super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.dtype, self.meta.qtype, self.tostring()))

    def __array_finalize__(self, obj):
        self.meta = MetaData() if obj is None else getattr(obj, 'meta', MetaData())



class QTemporalList(QList):
    '''An array object represents a q vector of datetime objects.'''

    def __getitem__(self, idx):
        return qtemporal(from_raw_qtemporal(numpy.ndarray.__getitem__(self, idx), -abs(self.meta.qtype)), qtype = -abs(self.meta.qtype))

    def __setitem__(self, idx, value):
        numpy.ndarray.__setitem__(self, idx, to_raw_qtemporal(value, - -abs(self.meta.qtype)))

    def raw(self, idx):
        '''Gets the raw representation of the datetime object at the specified 
        index.
        
           >>> t = qlist(numpy.array([366, 121, qnull(QDATE)]), qtype=QDATE_LIST)
           >>> print(t[0])
           2001-01-01 [metadata(qtype=-14)]
           >>> print(t.raw(0))
           366
        
        :Parameters:
         - `idx` (`integer`) - array index of the datetime object to be retrieved
         
        :returns: raw representation of the datetime object
        '''
        return numpy.ndarray.__getitem__(self, idx)

    def __repr__(self):
        return 'QTemporalList({}, qtype={})'.format(numpy.array2string(self, separator=', ', formatter={"int": QTemporal.__str__}), -abs(self.meta.qtype))

    def __str__(self):
        return numpy.array2string(self, separator=', ', formatter={"int": QTemporal.__str__})


def get_list_qtype(array):
    '''Finds out a corresponding qtype for a specified `QList`/`numpy.ndarray` 
    instance.
    
    :Parameters:
     - `array` (`QList` or `numpy.ndarray`) - array to be checked
    
    :returns: `integer` - qtype matching the specified array object
    '''
    if not isinstance(array, numpy.ndarray):
        raise ValueError('array parameter is expected to be of type: numpy.ndarray, got: %s' % type(array))

    if isinstance(array, QList):
        return -abs(array.meta.qtype)

    qtype = None

    if str(array.dtype) in ('|S1', '<U1', '>U1', '|U1') :
        qtype = QCHAR

    if qtype is None:
        qtype = Q_TYPE.get(array.dtype.type, None)

    if qtype is None and array.dtype.type in (numpy.datetime64, numpy.timedelta64):
        qtype = TEMPORAL_PY_TYPE.get(str(array.dtype), None)

    if qtype is None:
        # determinate type based on first element of the numpy array
        qtype = Q_TYPE.get(type(array[0]), QGENERAL_LIST)

    return qtype



def qlist(array, adjust_dtype = True, **meta):
    '''Converts an input array to q vector and enriches object instance with 
    meta data.

    Returns a :class:`.QList` instance for non-datetime vectors. For datetime 
    vectors :class:`.QTemporalList` is returned instead.

    If parameter `adjust_dtype` is `True` and q type retrieved via 
    :func:`.get_list_qtype` doesn't match one provided as a `qtype` parameter 
    guessed q type, underlying numpy.array is converted to correct data type.
    
    `qPython` internally represents ``(0x01;0x02;0xff)`` q list as:
    ``<class 'qpython.qcollection.QList'> dtype: int8 qtype: -4: [ 1  2 -1]``.
    This object can be created by calling the :func:`.qlist` with following 
    arguments:
    
    - `byte numpy.array`:
       
       >>> v = qlist(numpy.array([0x01, 0x02, 0xff], dtype=numpy.byte))
       >>> print('%s dtype: %s qtype: %d: %s' % (type(v), v.dtype, v.meta.qtype, v))
       <class 'qpython.qcollection.QList'> dtype: int8 qtype: -4: [ 1  2 -1]
    
    - `int32 numpy.array` with explicit conversion to `QBYTE_LIST`:   
      
       >>> v = qlist(numpy.array([1, 2, -1]), qtype = QBYTE_LIST)
       >>> print('%s dtype: %s qtype: %d: %s' % (type(v), v.dtype, v.meta.qtype, v))
       <class 'qpython.qcollection.QList'> dtype: int8 qtype: -4: [ 1  2 -1]
    
    - plain Python `integer` list with explicit conversion to `QBYTE_LIST`:   
       
       >>> v = qlist([1, 2, -1], qtype = QBYTE_LIST)
       >>> print('%s dtype: %s qtype: %d: %s' % (type(v), v.dtype, v.meta.qtype, v))
       <class 'qpython.qcollection.QList'> dtype: int8 qtype: -4: [ 1  2 -1]

    - numpy datetime64 array with implicit conversion to `QDATE_LIST`:   
       
       >>> v = qlist(numpy.array([numpy.datetime64('2001-01-01'), numpy.datetime64('2000-05-01'), numpy.datetime64('NaT')], dtype='datetime64[D]'))
       >>> print('%s dtype: %s qtype: %d: %s' % (type(v), v.dtype, v.meta.qtype, v))
       <class 'qpython.qcollection.QList'> dtype: datetime64[D] qtype: -14: ['2001-01-01' '2000-05-01' 'NaT']
       
    - numpy datetime64 array with explicit conversion to `QDATE_LIST`:   
       
       >>> v = qlist(numpy.array([numpy.datetime64('2001-01-01'), numpy.datetime64('2000-05-01'), numpy.datetime64('NaT')], dtype='datetime64[D]'), qtype = QDATE_LIST)
       >>> print('%s dtype: %s qtype: %d: %s' % (type(v), v.dtype, v.meta.qtype, v))
       <class 'qpython.qcollection.QList'> dtype: datetime64[D] qtype: -14: ['2001-01-01' '2000-05-01' 'NaT']

    
    :Parameters:
     - `array` (`tuple`, `list`, `numpy.array`) - input array to be converted
     - `adjust_dtype` (`boolean`) - determine whether data type of vector should
       be adjusted if it doesn't match default representation. **Default**: ``True``
       
     .. note:: numpy `datetime64` and `timedelta64` arrays are not converted
               to raw temporal vectors if `adjust_dtype` is ``True``
    
    :Kwargs:
     - `qtype` (`integer` or `None`) - qtype indicator
     
    :returns: `QList` or `QTemporalList` - array representation of the list
    
    :raises: `ValueError` 
    '''
    if type(array) in (list, tuple):
        if meta and 'qtype' in meta and meta['qtype'] == QGENERAL_LIST:
            # force shape and dtype for generic lists
            tarray = numpy.ndarray(shape = len(array), dtype = numpy.dtype('O'))
            for i in range(len(array)):
                tarray[i] = array[i]
            array = tarray
        else:
            array = numpy.array(array)

    if not isinstance(array, numpy.ndarray):
        raise ValueError('array parameter is expected to be of type: numpy.ndarray, list or tuple. Was: %s' % type(array))

    qtype = None
    is_numpy_temporal = array.dtype.type in (numpy.datetime64, numpy.timedelta64)

    if meta and 'qtype' in meta:
        qtype = -abs(meta['qtype'])
        dtype = PY_TYPE[qtype]
        if adjust_dtype and dtype != array.dtype and not is_numpy_temporal:
            array = array.astype(dtype = dtype)

    qtype = get_list_qtype(array) if qtype is None else qtype
    meta['qtype'] = qtype

    # With new QTemporal system, temporal lists use regular QList like integers
    vector = array.view(QList)
    vector._meta_init(**meta)
    return vector



class QDictionary(object):
    '''Represents a q dictionary.
    
    Dictionary examples:
    
    >>> # q: 1 2!`abc`cdefgh
    >>> print(QDictionary(qlist(numpy.array([1, 2], dtype=numpy.int64), qtype=QLONG_LIST), 
    ...                    qlist(numpy.array(['abc', 'cdefgh']), qtype = QSYMBOL_LIST)))
    [1 2]!['abc' 'cdefgh']
       
    >>> # q: (1;2h;3.234;"4")!(`one;2 3;"456";(7;8 9))
    >>> print(QDictionary([numpy.int64(1), numpy.int16(2), numpy.float64(3.234), '4'], 
    ...                    [numpy.bytes_('one'), qlist(numpy.array([2, 3]), qtype=QLONG_LIST), '456', [numpy.int64(7), qlist(numpy.array([8, 9]), qtype=QLONG_LIST)]]))
    [1, 2, 3.234, '4']!['one', QList([2, 3], dtype=int64), '456', [7, QList([8, 9], dtype=int64)]]
    
    :Parameters:
     - `keys` (`QList`, `tuple` or `list`) - dictionary keys
     - `values` (`QList`, `QTable`, `tuple` or `list`) - dictionary values
    '''
    def __init__(self, keys, values):
        if not isinstance(keys, (QList, tuple, list, numpy.ndarray)):
            raise ValueError('%s expects keys to be of type: QList, tuple or list. Actual type: %s' % (self.__class__.__name__, type(keys)))
        if not isinstance(values, (QTable, QList, tuple, list, numpy.ndarray)):
            raise ValueError('%s expects values to be of type: QTable, QList, tuple or list. Actual type: %s' % (self.__class__.__name__, type(values)))
        if len(keys) != len(values):
            raise ValueError('Number of keys: %d doesn`t match number of values: %d' % (len(keys), len(values)))

        self.keys = keys
        self.values = values

    def __str__(self, *args, **kwargs):
        return '%s!%s' % (self.keys, self.values)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        idx = 0
        for key in self.keys:
            if key != other.keys[idx] or self.values[idx] != other.values[idx]:
                return False
            idx += 1

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def _find_key_(self, key):
        idx = 0
        for k in self.keys:
            if key == k:
                return idx
            idx += 1

        raise KeyError('QDictionary doesn`t contain key: %s' % key)

    def __getitem__(self, key):
        return self.values[self._find_key_(key)]

    def __setitem__(self, key, value):
        self.values[self._find_key_(key)] = value

    def __len__(self):
        return len(self.keys)

    def __iter__(self):
        return iter(self.keys)

    def items(self):
        '''Return a copy of the dictionary's list of ``(key, value)`` pairs.'''
        return [(self.keys[x], self.values[x]) for x in range(len(self.keys))]

    def iteritems(self):
        '''Return an iterator over the dictionary's ``(key, value)`` pairs.'''
        for x in range(len(self.keys)):
            yield (self.keys[x], self.values[x])

    def iterkeys(self):
        '''Return an iterator over the dictionary's keys.'''
        return iter(self.keys)

    def itervalues(self):
        '''Return an iterator over the dictionary's values.'''
        return iter(self.values)



class QTable(numpy.recarray):
    '''Represents a q table.
    
    Internal table data is stored as a `numpy.array` separately for each column.
    This mimics the internal representation of tables in q.
    '''
    def _meta_init(self, **meta):
        self.meta = MetaData(**meta)

    def __eq__(self, other):
        return numpy.array_equal(self, other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __array_finalize__(self, obj):
        self.meta = MetaData() if obj is None else getattr(obj, 'meta', MetaData())



def qtable(columns, data, **meta):
    '''Creates a QTable out of given column names and data, and initialises the 
    meta data.
    
    :class:`.QTable` is represented internally by `numpy.core.records.recarray`.
    Data for each column is converted to :class:`.QList` via :func:`.qlist` 
    function. If qtype indicator is defined for a column, this information
    is used for explicit array conversion.
    
    Table examples:
  
      >>> # q: flip `name`iq!(`Dent`Beeblebrox`Prefect;98 42 126)
      >>> t = qtable(qlist(numpy.array(['name', 'iq']), qtype = QSYMBOL_LIST), 
      ...     [qlist(numpy.array(['Dent', 'Beeblebrox', 'Prefect'])), 
      ...      qlist(numpy.array([98, 42, 126], dtype=numpy.int64))])
      >>> print('%s dtype: %s meta: %s: %s' % (type(t), t.dtype, t.meta, t))
      <class 'qpython.qcollection.QTable'> dtype: [('name', 'S10'), ('iq', '<i8')] meta: metadata(iq=-7, qtype=98, name=-11): [('Dent', 98L) ('Beeblebrox', 42L) ('Prefect', 126L)]
      
      >>> # q: flip `name`iq!(`Dent`Beeblebrox`Prefect;98 42 126)
      >>> t = qtable(qlist(numpy.array(['name', 'iq']), qtype = QSYMBOL_LIST),
      ...           [qlist(['Dent', 'Beeblebrox', 'Prefect'], qtype = QSYMBOL_LIST), 
      ...            qlist([98, 42, 126], qtype = QLONG_LIST)])
      >>> print('%s dtype: %s meta: %s: %s' % (type(t), t.dtype, t.meta, t))
      <class 'qpython.qcollection.QTable'> dtype: [('name', 'S10'), ('iq', '<i8')] meta: metadata(iq=-7, qtype=98, name=-11): [('Dent', 98L) ('Beeblebrox', 42L) ('Prefect', 126L)]
      
      >>> # q: flip `name`iq!(`Dent`Beeblebrox`Prefect;98 42 126)
      >>> t = qtable(['name', 'iq'],
      ...            [['Dent', 'Beeblebrox', 'Prefect'], 
      ...             [98, 42, 126]],
      ...            name = QSYMBOL, iq = QLONG)
      >>> print('%s dtype: %s meta: %s: %s' % (type(t), t.dtype, t.meta, t)) 
      <class 'qpython.qcollection.QTable'> dtype: [('name', 'S10'), ('iq', '<i8')] meta: metadata(iq=-7, qtype=98, name=-11): [('Dent', 98L) ('Beeblebrox', 42L) ('Prefect', 126L)]
      
      >>> # q: flip `name`iq`fullname!(`Dent`Beeblebrox`Prefect;98 42 126;("Arthur Dent"; "Zaphod Beeblebrox"; "Ford Prefect"))
      >>> t = qtable(('name', 'iq', 'fullname'),
      ...            [qlist(numpy.array(['Dent', 'Beeblebrox', 'Prefect']), qtype = QSYMBOL_LIST), 
      ...             qlist(numpy.array([98, 42, 126]), qtype = QLONG_LIST),
      ...             qlist(numpy.array(["Arthur Dent", "Zaphod Beeblebrox", "Ford Prefect"]), qtype = QSTRING_LIST)])
      <class 'qpython.qcollection.QTable'> dtype: [('name', 'S10'), ('iq', '<i8'), ('fullname', 'O')] meta: metadata(iq=-7, fullname=0, qtype=98, name=-11): [('Dent', 98L, 'Arthur Dent') ('Beeblebrox', 42L, 'Zaphod Beeblebrox') ('Prefect', 126L, 'Ford Prefect')]
    
    :Parameters:
     - `columns` (list of `strings`) - table column names 
     - `data` (list of lists) - list of columns containing table data
    
    :Kwargs:
     - `meta` (`integer`) - qtype for particular column 
    
    :returns: `QTable` - representation of q table
    
    :raises: `ValueError`
    '''
    if len(columns) != len(data):
        raise ValueError('Number of columns doesn`t match the data layout. %s vs %s' % (len(columns), len(data)))

    meta = {} if not meta else meta

    if not 'qtype' in meta:
        meta['qtype'] = QTABLE

    dtypes = []
    for i in range(len(columns)):
        column_name = columns[i] if isinstance(columns[i], str) else columns[i].decode("utf-8")
        
        if isinstance(data[i], str):
            # convert character list (represented as string) to numpy representation
            data[i] = numpy.array(list(data[i]), dtype = numpy.bytes_)
        if isinstance(data[i], bytes):
            data[i] = numpy.array(list(data[i].decode()), dtype = numpy.bytes_)

        if column_name in meta:
            data[i] = qlist(data[i], qtype = meta[column_name])
        elif not isinstance(data[i], QList):
            if type(data[i]) in (list, tuple):
                data[i] = qlist(data[i], qtype = QGENERAL_LIST)
            else:
                data[i] = qlist(data[i])

        
        meta[column_name] = data[i].meta.qtype
        dtypes.append((column_name, data[i].dtype))

    table = numpy.core.records.fromarrays(data, dtype = dtypes)
    table = table.view(QTable)

    table._meta_init(**meta)
    return table



class QKeyedTable(object):
    '''Represents a q keyed table.
    
    :class:`.QKeyedTable` is built with two :class:`.QTable`s, one representing
    keys and the other values.
    
    Keyed tables example:
    
        >>> # q: ([eid:1001 1002 1003] pos:`d1`d2`d3;dates:(2001.01.01;2000.05.01;0Nd))
        >>> t = QKeyedTable(qtable(['eid'],
        ...                [qlist(numpy.array([1001, 1002, 1003]), qtype = QLONG_LIST)]),
        ...         qtable(['pos', 'dates'],
        ...                [qlist(numpy.array(['d1', 'd2', 'd3']), qtype = QSYMBOL_LIST), 
        ...                 qlist(numpy.array([366, 121, qnull(QDATE)]), qtype = QDATE_LIST)]))
        >>> print('%s: %s' % (type(t), t))
        >>> print('%s dtype: %s meta: %s' % (type(t.keys), t.keys.dtype, t.keys.meta))
        >>> print('%s dtype: %s meta: %s' % (type(t.values), t.values.dtype, t.values.meta))
        <class 'qpython.qcollection.QKeyedTable'>: [(1001L,) (1002L,) (1003L,)]![('d1', 366) ('d2', 121) ('d3', -2147483648)]
        <class 'qpython.qcollection.QTable'> dtype: [('eid', '<i8')] meta: metadata(qtype=98, eid=-7)
        <class 'qpython.qcollection.QTable'> dtype: [('pos', 'S2'), ('dates', '<i4')] meta: metadata(dates=-14, qtype=98, pos=-11)
    
    :Parameters:
     - `keys` (`QTable`) - table keys
     - `values` (`QTable`) - table values
    
    :raises: `ValueError`
    '''
    def __init__(self, keys, values):
        if not isinstance(keys, QTable):
            raise ValueError('Keys array is required to be of type: QTable')

        if not isinstance(values, QTable):
            raise ValueError('Values array is required to be of type: QTable')

        if len(keys) != len(values):
            raise ValueError('Keys and value arrays cannot have different length')
        self.keys = keys
        self.values = values

    def __str__(self, *args, **kwargs):
        return '%s!%s' % (self.keys, self.values)

    def __eq__(self, other):
        return isinstance(other, QKeyedTable) and numpy.array_equal(self.keys, other.keys) and numpy.array_equal(self.values, other.values)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.keys)

    def __iter__(self):
        return iter(self.keys)

    def items(self):
        '''Return a copy of the keyed table's list of ``(key, value)`` pairs.'''
        return [(self.keys[x], self.values[x]) for x in range(len(self.keys))]

    def iteritems(self):
        '''Return an iterator over the keyed table's ``(key, value)`` pairs.'''
        for x in range(len(self.keys)):
            yield (self.keys[x], self.values[x])

    def iterkeys(self):
        '''Return an iterator over the keyed table's keys.'''
        return iter(self.keys)

    def itervalues(self):
        '''Return an iterator over the keyed table's values.'''
        return iter(self.values)



"""
Q/kdb+ connection utilities for qmcp
Clean, minimal interface for connecting to and querying q servers
"""

from .qpython.qconnection import QConnection
from .qpython.qmetaconversion import MetaData
from .qpython.qtype import Char, String
import os
import time
import socket
from sugar import dmap, spl
import pandas as pd
import numpy as np


def _get_hostname():
    """Get current hostname"""
    return socket.gethostname()


class _QDecodeWrapper:
    """Wrapper for QConnection that handles pandas DataFrame encoding/decoding"""
    
    def __init__(self, q):
        self.q = q
        
    def decode(self, result):
        """
        Decode q types in pandas DataFrame results and iterables with type preservation
        
        Args:
            result: Result from q query (DataFrame, dict, list, etc.)
            
        Returns:
            Result with q types properly wrapped (Char/String) or preserved (symbols)
        """
        return result
        # Handle pandas DataFrames with meta information
        if type(result) == pd.DataFrame:
            rd = result.meta.as_dict()
            
            # Handle regular columns
            for col in rd:
                qtype = rd[col]
                if qtype in (0, 10, 11) and col in result.columns:  # string, char, or symbol columns
                    try:
                        if qtype == 0:  # string
                            result[col] = result[col].map(dmap(lambda x: String(x), result[col].drop_duplicates()))
                        elif qtype == 10:  # char  
                            result[col] = result[col].map(dmap(lambda x: Char(x), result[col].drop_duplicates()))
                        elif qtype == 11:  # symbol: decode bytes to plain strings
                            result[col] = result[col].map(dmap(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x, result[col].drop_duplicates()))
                    except:
                        pass
            
            # Handle simple index
            for col in rd:
                qtype = rd[col]
                if qtype in (0, 10, 11) and result.index.name == col and hasattr(result.index, '__iter__'):
                    try:
                        if qtype == 0:  # string
                            decoded_index = [String(x) for x in result.index]
                            result.index = pd.Index(decoded_index, name=result.index.name)
                        elif qtype == 10:  # char
                            decoded_index = [Char(x) for x in result.index]
                            result.index = pd.Index(decoded_index, name=result.index.name)
                        elif qtype == 11:  # symbol: decode bytes to plain strings
                            decoded_index = [x.decode('utf-8') if isinstance(x, bytes) else x for x in result.index]
                            result.index = pd.Index(decoded_index, name=result.index.name)
                    except:
                        pass
            
            # Handle MultiIndex
            if isinstance(result.index, pd.MultiIndex):
                try:
                    new_levels = []
                    for i, level in enumerate(result.index.levels):
                        level_name = result.index.names[i]
                        if level_name in rd:
                            qtype = rd[level_name]
                            if qtype == 0:  # string
                                try:
                                    decoded_level = [String(x) for x in level]
                                    new_levels.append(pd.Index(decoded_level))
                                except:
                                    new_levels.append(level)
                            elif qtype == 10:  # char
                                try:
                                    decoded_level = [Char(x) for x in level]
                                    new_levels.append(pd.Index(decoded_level))
                                except:
                                    new_levels.append(level)
                            elif qtype == 11:  # symbol: decode bytes to plain strings
                                try:
                                    decoded_level = [x.decode('utf-8') if isinstance(x, bytes) else x for x in level]
                                    new_levels.append(pd.Index(decoded_level))
                                except:
                                    new_levels.append(level)
                            else:  # other types: keep as-is
                                new_levels.append(level)
                        else:
                            new_levels.append(level)
                    
                    # Reconstruct MultiIndex with type-wrapped levels
                    result.index = pd.MultiIndex(
                        levels=new_levels,
                        codes=result.index.codes,
                        names=result.index.names
                    )
                except:
                    pass
        
        # Handle iterables (lists, tuples, etc.) - decode bytes to strings recursively
        elif hasattr(result, '__iter__') and not isinstance(result, (str, bytes)):
            try:
                if isinstance(result, dict):
                    # Recursively decode dictionary values
                    return {k: self.decode(v) for k, v in result.items()}
                else:
                    # Handle lists, tuples, etc.
                    decoded_items = []
                    for item in result:
                        if isinstance(item, bytes):
                            # Decode bytes to plain strings (no type wrapper for generic lists)
                            decoded_items.append(item.decode('utf-8'))
                        else:
                            # Recursively decode nested structures
                            decoded_items.append(self.decode(item))
                    
                    # Return same type as input (list, tuple, etc.)
                    return type(result)(decoded_items)
            except:
                pass
                    
        return result

    def __call__(self, *args, **kwargs):
        # Handle DataFrame arguments
        for arg in args:
            if isinstance(arg, pd.DataFrame):
                if not hasattr(arg, 'meta'):
                    arg.meta = MetaData(qtype=98)
                    
                # Set date column metadata
                for c in 'd, date'-spl:
                    if (c in arg.columns and 
                        isinstance(arg[c].dtype, np.dtype) and 
                        arg[c].dtype != np.dtype(np.object_) and
                        c not in arg.meta.__dict__):
                        arg.meta[c] = 14
                        
                # Handle string columns
                if 'strcols' in kwargs:
                    for sc in kwargs['strcols']-spl:
                        arg.meta[sc] = 0
                        
        # Execute query
        r = self.q(*args, **kwargs)
        
        # Decode using the factored out method
        return self.decode(r)


def connect_to_q(connection_string, connection_timeout=5):
    """
    Connect to q server using a connection string

    Args:
        connection_string: Connection string in format 'host:port' or 'host:port:user:password'
        connection_timeout: seconds to wait for connection (default 5)

    Returns:
        _QDecodeWrapper instance

    Note: Connection logic and config management is handled by qcomms.connect_to_q,
          which calls this function with a fully-formed connection string.
    """
    if not connection_string:
        raise ValueError("Connection string required")

    return _qConnect(str(connection_string), pandas=True, connection_timeout=connection_timeout)


def _qConnect(qCredentials, pandas, connection_timeout=5):
    """
    Connect to q server with socket timeout
    
    Args:
        qCredentials: 'host:port' or 'host:port:user:passwd'
        pandas: return pandas-enabled connection
        connection_timeout: socket timeout in seconds
        
    Returns:
        QConnection or _QDecodeWrapper
    """
    qCreds = tuple(qCredentials.split(':'))
    host, port, user, passwd = qCreds if len(qCreds) == 4 else (qCreds + (None, None))
    port = int(port)
    
    if host == _get_hostname():
        host = 'localhost'
        
    # Create connection with socket timeout
    q = QConnection(host, port, user, passwd, pandas=pandas)
    
    # Set socket timeout before opening connection
    import socket
    original_timeout = socket.getdefaulttimeout()
    socket.setdefaulttimeout(connection_timeout)
    
    try:
        q.open()
        return _QDecodeWrapper(q) if pandas else q
    finally:
        # Restore original socket timeout
        socket.setdefaulttimeout(original_timeout)


import json
import os
from pathlib import Path
from typing import Dict, Any


def get_docs_path() -> Path:
    """Get the path to the qython documentation directory"""
    return Path(__file__).parent / "docs"


def get_documentation(topic: str = "home") -> str:
    """Get Qython documentation by topic
    
    Args:
        topic: Documentation topic filename (without extension)
        
    Returns:
        Markdown documentation content
        
    Raises:
        ValueError: If topic contains invalid characters or path traversal
        FileNotFoundError: If documentation file doesn't exist
    """
    # Security: prevent path traversal
    if ".." in topic or "/" in topic or "\\" in topic:
        raise ValueError(f"Invalid topic name: {topic}")
    
    # Only allow alphanumeric and basic characters
    if not topic.replace("_", "").replace("-", "").isalnum():
        raise ValueError(f"Invalid topic name: {topic}")
    
    docs_path = get_docs_path()
    
    # Special handling for "home" topic - combine multiple files
    if topic == "home":
        home_files = ["home.md", "home_numpy.md", "home_db.md"]
        contents = []
        
        for filename in home_files:
            file_path = docs_path / filename
            if file_path.exists():
                # Ensure the resolved path is still within the docs directory
                if not str(file_path.resolve()).startswith(str(docs_path.resolve())):
                    raise ValueError(f"Invalid file path: {filename}")
                
                try:
                    with open(file_path, 'r', encoding='utf-8') as f:
                        contents.append(f.read())
                except OSError as e:
                    raise ValueError(f"Could not read documentation file '{filename}': {e}")
        
        if not contents:
            raise FileNotFoundError(f"No home documentation files found")
        
        return '\n\n'.join(contents)
    
    # Regular handling for other topics
    md_file = docs_path / f"{topic}.md"
    
    if not md_file.exists():
        # List available topics
        available = [f.stem for f in docs_path.glob("*.md")]
        raise FileNotFoundError(f"Documentation topic '{topic}' not found. Available: {', '.join(available)}")
    
    # Ensure the resolved path is still within the docs directory (extra security)
    if not str(md_file.resolve()).startswith(str(docs_path.resolve())):
        raise ValueError(f"Invalid topic path: {topic}")
    
    try:
        with open(md_file, 'r', encoding='utf-8') as f:
            return f.read()
    except OSError as e:
        raise ValueError(f"Could not read documentation file '{topic}': {e}")

# def get_documentation_json(topic: str = "home") -> Dict[str, Any]:
#     """Get Qython documentation by topic (JSON format)
#     
#     Args:
#         topic: Documentation topic filename (without extension)
#         
#     Returns:
#         Dictionary containing the documentation content
#         
#     Raises:
#         ValueError: If topic contains invalid characters or path traversal
#         FileNotFoundError: If documentation file doesn't exist
#     """
#     # Security: prevent path traversal
#     if ".." in topic or "/" in topic or "\\" in topic:
#         raise ValueError(f"Invalid topic name: {topic}")
#     
#     # Only allow alphanumeric and basic characters
#     if not topic.replace("_", "").replace("-", "").isalnum():
#         raise ValueError(f"Invalid topic name: {topic}")
#     
#     docs_path = get_docs_path()
#     json_file = docs_path / f"{topic}.json"
#     
#     if not json_file.exists():
#         # List available topics
#         available = [f.stem for f in docs_path.glob("*.json")]
#         raise FileNotFoundError(f"Documentation topic '{topic}' not found. Available: {', '.join(available)}")
#     
#     # Ensure the resolved path is still within the docs directory (extra security)
#     if not str(json_file.resolve()).startswith(str(docs_path.resolve())):
#         raise ValueError(f"Invalid topic path: {topic}")
#     
#     try:
#         with open(json_file, 'r', encoding='utf-8') as f:
#             return json.load(f)
#     except json.JSONDecodeError as e:
#         raise ValueError(f"Invalid JSON in documentation file '{topic}': {e}")


def list_available_topics() -> Dict[str, str]:
    """List all available documentation topics
    
    Returns:
        Dictionary mapping topic names to their titles
    """
    docs_path = get_docs_path()
    topics = {}
    
    for json_file in docs_path.glob("*.json"):
        topic_name = json_file.stem
        try:
            with open(json_file, 'r', encoding='utf-8') as f:
                data = json.load(f)
                title = data.get('title', topic_name)
                topics[topic_name] = title
        except (json.JSONDecodeError, OSError):
            # Skip invalid files
            continue
    
    return topics


#!/usr/bin/env python3
"""
qmcp Server - MCP Server for q/kdb+ integration

A Model Context Protocol server that provides q/kdb+ connectivity
with flexible connection management and query execution.
"""

from mcp.server.fastmcp import FastMCP
import os
import sys
from typing import Literal, Optional

from . import qcomms
from . import translate
from . import qythonhelp
from . import auth
from .config import get_config_path, load_config, get_config

# Initialize the MCP server
mcp = FastMCP("qmcp")

@mcp.tool(structured_output=False)
def MultiTool(
    action: Literal[
        "get_config_file_path", "reload_config",
        "interrupt_current_query", "wait_for_task_result", "get_last_result_console_view",
        "setup_qython_namespace", "export_qython_namespace_file",
        "request_api_key", "qython_help",
        "list_tables", "describe_table"
    ],
    parameter = None
) -> str:
    """Admin, help, and introspection utilities. The following utilities require a parameter:
-request_api_key: email address
-export_qython_namespace_file: output file path, default "qython.q"
-qython_help: Qython help page, default "home". You should run this when asked to code in the Qython language
-describe_table: table name
-wait_for_task_result: seconds to wait as integer
Use reload_config tool after a change to the config file to apply the new changes.
Before running Qython code, you must use the setup_qython_namespace tool."""

    if action == "get_config_file_path":
        return str(get_config_path())

    elif action == "reload_config":
        try:
            import qmcp.config
            qmcp.config._config = None
            load_config()
            return "✅ Configuration reloaded successfully"
        except Exception as e:
            return f"❌ Failed to reload configuration: {e}"

    elif action == "interrupt_current_query":
        return qcomms.interrupt_current_query()

    elif action == "wait_for_task_result":
        return qcomms.get_current_task_result(wait_seconds=parameter)

    elif action == "get_last_result_console_view":
        return qcomms.get_last_query_result_q_view()

    elif action == "setup_qython_namespace":
        return translate.setup_qython_namespace()

    elif action == "export_qython_namespace_file":
        return translate.export_qython_namespace(file_path=parameter or 'qython.q')

    elif action == "request_api_key":
        if not parameter:
            return "❌ email address required for request_api_key action"
        return auth.request_api_key(email_address=parameter)

    elif action == "qython_help":
        return qythonhelp.qython_help(topic=parameter or 'home')

    elif action == "list_tables":
        return qcomms.list_tables()

    elif action == "describe_table":
        if not parameter:
            return "❌ table name required for describe_table action"
        return qcomms.describe_table(table=parameter)

    else:
        return f"❌ Unknown action: {action}"

# Q connection and query tools
connect_to_q = mcp.tool(structured_output=False)(qcomms.connect_to_q)
query_q = mcp.tool(structured_output=False)(qcomms.query_q)

# Translation tools
translate_qython_to_q = mcp.tool(structured_output=False)(translate.translate_qython_to_q)
translate_q_to_qython = mcp.tool(structured_output=False)(translate.translate_q_to_qython)
translate_and_run_qython = mcp.tool(structured_output=False)(translate.translate_and_run_qython)
translate_qython_file_to_q_file = mcp.tool(structured_output=False)(translate.translate_qython_file_to_q_file)
translate_qython_to_q_file = mcp.tool(structured_output=False)(translate.translate_qython_to_q_file)
translate_qython_file_to_q = mcp.tool(structured_output=False)(translate.translate_qython_file_to_q)
run_qython_file_via_IPC = mcp.tool(structured_output=False)(translate.run_qython_file_via_IPC)
run_q_file_via_IPC = mcp.tool(structured_output=False)(translate.run_q_file_via_IPC)

def main():
    """Main entry point for the MCP server"""
    # Load config early to apply any fallbacks and set warnings
    get_config()
    mcp.run()


if __name__ == "__main__":
    main()


from qmcp.qlib import connect_to_q
from typing import List as ListType, Union, Any, Tuple
from .disambiguate import disambiguate_step
# from callLLM import call_llm
import os

package_dir = os.path.dirname(__file__)
parseq_q_file = os.path.join(package_dir, 'parseq_ns.q')
with open(parseq_q_file, 'r') as f:
    parseq_q_code = f.read()

def quote(x):
    escaped = str(x).replace('\\', '\\\\').replace('"', '\\"')
    return f'"{escaped}"'

glyph_map = {
    '@': 'at', '!': 'bang', '::': 'colon_colon',
    '.': 'dot', '$': 'dollar', '#': 'hash', '?': 'query',
    '_': 'underscore', ',': 'comma', '/': 'slash',
    '\\': 'backslash'
    }


# AST Node classes
class ASTNode:
    def __init__(self, node_type, value=None, children=None, eager=False, is_inside_dict = False):
        self.node_type = node_type
        self.value = value
        self.children = children or []
        self.eager = eager
        self.is_inside_dict = is_inside_dict

# Helper functions
def print_node_tree(node, indent=0):
    """Print AST node tree with node_type and eager status"""
    prefix = "  " * indent
    eager_status = f" (eager={node.eager})" if hasattr(node, 'eager') else ""
    node_type_status = f" (node_type={node.node_type})" if hasattr(node, 'node_type') else " (NO NODE_TYPE)"
    
    if node.node_type == 'variable':
        print(f"{prefix}Variable: {node.value}{eager_status}{node_type_status}")
    elif node.node_type == 'symbol':
        print(f"{prefix}Symbol: `{node.value}{eager_status}{node_type_status}")
    elif node.node_type == 'integer':
        print(f"{prefix}Integer: {node.value}{eager_status}{node_type_status}")
    elif node.node_type == 'float':
        print(f"{prefix}Float: {node.value}{eager_status}{node_type_status}")
    elif node.node_type == 'boolean':
        print(f"{prefix}Boolean: {node.value}{eager_status}{node_type_status}")
    elif node.node_type == 'string':
        print(f"{prefix}String: {node.value}{eager_status}{node_type_status}")
    elif node.node_type == 'char':
        print(f"{prefix}Char: '{node.value}'{eager_status}{node_type_status}")
    elif node.node_type == 'sequence':
        print(f"{prefix}Sequence{eager_status}{node_type_status}:")
        for i, child in enumerate(node.children):
            print(f"{prefix}  Statement {i}:")
            print_node_tree(child, indent + 2)
    elif node.node_type == 'function':
        print(f"{prefix}Function: {node.value}{eager_status}{node_type_status}")
    elif node.node_type == 'symbol_list':
        print(f"{prefix}SymbolList{eager_status}{node_type_status}:")
        for i, child in enumerate(node.children):
            print_node_tree(child, indent + 1)
    elif node.node_type == 'lambda':
        param_names = getattr(node, 'param_names', [])
        print(f"{prefix}Lambda{eager_status}{node_type_status}:")
        print(f"{prefix}  Params: {param_names}")
        print(f"{prefix}  Body:")
        print_node_tree(node.children[0], indent + 2)
    elif node.node_type == 'list':
        print(f"{prefix}List{eager_status}{node_type_status}:")
        for i, child in enumerate(node.children):
            print(f"{prefix}  [{i}]:")
            print_node_tree(child, indent + 2)
    elif node.node_type == 'dict':
        print(f"{prefix}Dict{eager_status}{node_type_status}:")
        print(f"{prefix}  Keys:")
        print_node_tree(node.children[0], indent + 2)
        print(f"{prefix}  Values:")
        print_node_tree(node.children[1], indent + 2)
    else:
        print(f"{prefix}Unknown ({node.node_type}): {node.value}{eager_status}{node_type_status}")
        if hasattr(node, 'children') and node.children:
            print(f"{prefix}  Children:")
            for i, child in enumerate(node.children):
                print(f"{prefix}    [{i}]:")
                print_node_tree(child, indent + 3)

def inspect_node(node):
    """Inspect a node to understand its structure"""
    print(f"Node type: {node.node_type}")
    print(f"Node value: {node.value}")
    print(f"Node children: {len(node.children) if node.children else 0}")
    if node.children:
        for i, child in enumerate(node.children):
            print(f"  Child {i}: {child.node_type} = {child.value}")
    print(f"Node attributes: {[attr for attr in dir(node) if not attr.startswith('_')]}")
    print(f"Node eager: {node.eager}")
    print()


# Tokenizer
def tokenize(text: str) -> ListType[str]:
    """Tokenize the LISP-like input into a list of tokens"""
    tokens = []
    i = 0
    while i < len(text):
        char = text[i]
        if char.isspace():
            i += 1
        elif char in '[](),{}':
            tokens.append(char)
            i += 1
        elif char == ':':
            # Handle dict syntax like LSymbol[s, t]:
            tokens.append(char)
            i += 1
        else:
            # Read a word/identifier
            start = i
            while i < len(text) and text[i] not in '[](){},:' and not text[i].isspace():
                i += 1
            tokens.append(text[start:i])
    return tokens

# Parser
class Parser:
    def __init__(self, tokens: ListType[str]):
        self.tokens = tokens
        self.pos = 0
    
    def current_token(self) -> str:
        if self.pos >= len(self.tokens):
            return None
        return self.tokens[self.pos]
    
    def consume(self) -> str:
        token = self.current_token()
        self.pos += 1
        return token
    
    def parse(self) -> ASTNode:
        """Parse the tokens into an AST"""
        return self.parse_expression()
    
    def parse_expression(self, idx = 0, is_inside_dict=False) -> ASTNode:
        token = self.current_token()
        
        if token == '[':
            return self.parse_list(is_inside_dict)
        elif token == '{':
            return self.parse_dict()
        elif token and token[0].isupper():
            # Type constructor like Symbol[name], Int[5], etc.
            return self.parse_type_constructor()
        else:
            # Simple token - consume it and parse as simple token
            token = self.consume()
            return self.parse_simple_token(token)
    
    def parse_list(self, is_inside_dict) -> ASTNode:
        """Parse a bracketed list [...]"""
        self.consume()  # consume '['
        elements = []
        idx = 0
        while self.current_token() != ']':
            if idx == 1 and not is_inside_dict:
                elements[0].eager = True
            if self.current_token() == ',':
                self.consume()  # skip comma
                continue
            parsed_result = self.parse_expression(idx, is_inside_dict)
            # Handle case where parsing returns symbol_list (needs flattening)
            if parsed_result.node_type == 'symbol_list':
                elements.extend(parsed_result.children)
            else:
                elements.append(parsed_result)
            idx += 1
        
        self.consume()  # consume ']'
        return ASTNode('list', children=elements)
    
    def parse_dict(self) -> ASTNode:
        """Parse a dictionary {...}"""
        self.consume()  # consume '{'
        
        # Parse keys
        keys = self.parse_expression(0, is_inside_dict = True)
        
        if self.current_token() != ':':
            raise ValueError(f"Expected ':' in dict, got {self.current_token()}")
        self.consume()  # consume ':'
        
        # Parse values
        values = self.parse_expression(0, is_inside_dict = True)
        
        # Expect '}'
        if self.current_token() != '}':
            raise ValueError(f"Expected '}}' in dict, got {self.current_token()}")
        self.consume()  # consume '}'
        
        return ASTNode('dict', children=[keys, values])
    
    def parse_type_constructor(self) -> ASTNode:
        """Parse type constructors like Symbol[name], Int[5], Bool[0]"""
        type_name = self.consume()
        
        if self.current_token() != '[':
            # Just a plain identifier
            return ASTNode('string', value=type_name)
        
        self.consume()  # consume '['
        
        if type_name == 'Symbol':
            name = self.consume()
            self.consume()  # consume ']'
            return ASTNode('variable', value=name)
        elif type_name in ['Int', 'Long']:
            value = int(self.consume())
            self.consume()  # consume ']'
            return ASTNode('integer', value=value)
        elif type_name in ['Real', 'Float']:
            value = float(self.consume())
            self.consume()  # consume ']'
            return ASTNode('float', value=value)
        elif type_name == 'Bool':
            value = self.consume() == '1'
            self.consume()  # consume ']'
            return ASTNode('boolean', value=value)
        elif type_name == 'Char':
            char = self.consume()
            self.consume()  # consume ']'
            if char == ';':
                # Special case: semicolon indicates statement sequence
                return ASTNode('sequence', value=';')
            else:
                # Regular character
                return ASTNode('char', value=char)
        elif type_name == 'Builtin':
            func_name = self.consume()
            self.consume()  # consume ']'
            return ASTNode('variable', value=func_name)  # Functions are variables
        elif type_name == 'LSymbol':
            # Parse LSymbol[a, b, c] as a special node containing symbols to be flattened
            symbols = []
            while self.current_token() != ']':
                if self.current_token() == ',':
                    self.consume()  # skip comma
                    continue
                symbols.append(ASTNode('symbol', value=self.consume()))
            self.consume()  # consume ']'
            return ASTNode('symbol_list', children=symbols)
        elif type_name == 'Lambda':
            # Parse Lambda[params, body] - handle params specially
            # Expect '[' for parameter list
            if self.current_token() != '[':
                raise ValueError(f"Expected '[' for Lambda params, got {self.current_token()}")
            self.consume()  # consume '['
            
            # Parse parameter names directly (not as expressions)
            param_names = []
            while self.current_token() != ']':
                if self.current_token() == ',':
                    self.consume()  # skip comma
                    continue
                param_names.append(self.consume())  # Just take the parameter name as string
            self.consume()  # consume ']'
            
            if self.current_token() != ',':
                raise ValueError(f"Expected ',' in Lambda, got {self.current_token()}")
            self.consume()  # consume ','
            
            body = self.parse_expression()
            self.consume()  # consume ']'
            
            # Create lambda node with param_names as a property
            lambda_node = ASTNode('lambda', children=[body])
            lambda_node.param_names = param_names  # Store params as property
            return lambda_node
        elif type_name.startswith('L') and len(type_name) > 1:
            # Parse L{type} as a typed list (LSymbol, LLong, LChar, etc.)
            base_type = type_name[1:]  # Remove 'L' prefix
            elements = []
            while self.current_token() != ']':
                if self.current_token() == ',':
                    self.consume()  # skip comma
                    continue
                token = self.consume()
                
                # Convert based on base type
                if base_type == 'Symbol':
                    elements.append(ASTNode('symbol', value=token))
                elif base_type in ['Long', 'Int']:
                    elements.append(ASTNode('integer', value=int(token)))
                elif base_type in ['Real', 'Float']:
                    elements.append(ASTNode('float', value=float(token)))
                elif base_type == 'Bool':
                    elements.append(ASTNode('boolean', value=token == '1'))
                elif base_type == 'Char':
                    elements.append(ASTNode('string', value=token))
                else:
                    # Unknown base type, treat as string
                    elements.append(ASTNode('string', value=token))
            
            self.consume()  # consume ']'
            return ASTNode('list', children=elements)
        else:
            # Unknown type, parse contents as string
            content = self.consume()
            self.consume()  # consume ']'
            return ASTNode('string', value=f"{type_name}[{content}]")
    
    def parse_simple_token(self, token: str) -> ASTNode:
        """Parse simple tokens that aren't type constructors"""
        if token.isdigit():
            return ASTNode('integer', value=int(token))
        elif token.replace('.', '').isdigit() and '.' in token:
            return ASTNode('float', value=float(token))
        else:
            return ASTNode('string', value=token)

# AST Flattener - converts nested calls to step-by-step assignments
class Flattener:
    def flatten(self, node: ASTNode) -> tuple[ListType[str], str]:
        self.temp_counter = 0
        self.func_counter = 0
        return self.flatten_ast(node)
    def flatten_ast(self, node: ASTNode, **kwargs) -> tuple[ListType[str], str]:
        """
        Flatten nested function calls into step-by-step assignments.
        Returns (statements, final_expression)
        """
        statements = []
        
        if node.node_type == 'variable':
            # Replace glyphs with readable names (since functions are variables)
            var_name = glyph_map.get(node.value, node.value)
            return statements, var_name
        elif node.node_type == 'symbol':
            return statements, f"`{node.value}"  # Only symbols get backticks
        elif node.node_type == 'integer':
            return statements, str(node.value)
        elif node.node_type == 'float':
            value_str = str(node.value)
            if '.' not in value_str:
                value_str += '.0'
            return statements, value_str
        elif node.node_type == 'boolean':
            return statements, 'True' if node.value else 'False'
        elif node.node_type == 'string':
            return statements, node.value
        elif node.node_type == 'char':
            return statements, f"'{node.value}'"
        elif node.node_type == 'function':
            # Replace glyphs with readable names
            func_name = glyph_map.get(node.value, node.value)
            
            # Function nodes should just return their name - arguments are handled at the list level
            return statements, func_name
            
        elif node.node_type == 'list':
            if node.children and node.children[0].node_type == 'sequence':
                # First element is sequence marker: treat as statement sequence
                statements_list = node.children[1:]  # Skip the sequence marker
                for stmt in statements_list:
                    stmt_stmts, stmt_expr = self.flatten_ast(stmt)
                    statements.extend(stmt_stmts)
                # Return the last expression (Q semantics)
                return statements, stmt_expr
            elif node.children and node.children[0].eager:
                # First element is eager: treat as function call
                func_node = node.children[0]
                args = node.children[1:]
                
                # Flatten the function expression first
                func_stmts, func_expr = self.flatten_ast(func_node)
                statements.extend(func_stmts)
                
                # Apply glyph mapping if the function is a simple string
                if func_node.node_type == 'string':
                    func_expr = glyph_map.get(func_expr, func_expr)
                
                # Special handling for assignments: colon with arity 2 (check BEFORE flattening args)
                if func_expr == ':' and len(args) == 2:
                    var_node = args[0]    # Still an AST node
                    value_node = args[1]  # Still an AST node
                    
                    if value_node.node_type == 'lambda':
                        # Handle lambda assignment specially
                        # print(f"DEBUG: Found lambda assignment to {var_node.value}")
                        # print(f"DEBUG: Lambda node details:")
                        # inspect_node(value_node)
                        var_name = var_node.value
                        # Process lambda with custom function name
                        lambda_stmts, lambda_expr = self.flatten_ast(value_node, func_name=var_name)
                        statements.extend(lambda_stmts)
                        # No need for assignment since function is already defined with the right name
                        # Assignments always return None (Q semantics)
                        return statements, None
                    else:
                        # Regular assignment - flatten both sides
                        var_stmts, var_expr = self.flatten_ast(var_node)
                        val_stmts, val_expr = self.flatten_ast(value_node)
                        statements.extend(var_stmts)
                        statements.extend(val_stmts)
                        statements.append(f"{var_expr} = {val_expr}")
                        # Assignments always return None (Q semantics)
                        return statements, None
                
                # For non-assignment cases, flatten arguments normally
                arg_exprs = []
                for arg in args:
                    arg_stmts, arg_expr = self.flatten_ast(arg)
                    statements.extend(arg_stmts)
                    # Handle symbol_list flattening
                    if isinstance(arg_expr, list):
                        arg_exprs.extend(arg_expr)
                    else:
                        arg_exprs.append(arg_expr)
                
                args_str = ', '.join(arg_exprs)
                
                # Special handling for arithmetic operators - use infix notation
                arithmetic_ops = {'+': '+', '-': '-', '*': '*', '%': '/'}
                if func_expr in arithmetic_ops and len(arg_exprs) == 2:
                    infix_op = arithmetic_ops[func_expr]
                    expr = f"{arg_exprs[0]} {infix_op} {arg_exprs[1]}"
                else:
                    expr = f"{func_expr}({args_str})"
                
                # Always create temp variable for function calls (we need proper evaluation order)
                self.temp_counter += 1
                temp_name = f"temp{self.temp_counter}"
                statements.append(f"{temp_name} = {expr}")
                return statements, temp_name
            else:
                # First element is lazy or empty list: treat as regular list
                elem_exprs = []
                for elem in node.children:
                    elem_stmts, elem_expr = self.flatten_ast(elem)
                    statements.extend(elem_stmts)
                    # Handle symbol_list flattening
                    if isinstance(elem_expr, list):
                        elem_exprs.extend(elem_expr)
                    else:
                        elem_exprs.append(elem_expr)
                elements_str = ', '.join(elem_exprs)
                return statements, f"[{elements_str}]"
                
        elif node.node_type == 'symbol_list':
            # Flatten symbol list into individual symbols
            symbols = [f"`{child.value}" for child in node.children]
            return statements, f"[{', '.join(symbols)}]"
                
        elif node.node_type == 'lambda':
            # Generate def function for lambda since Python doesn't support multi-line lambdas
            body_node = node.children[0]  # Body is now the only child
            
            # Generate function name - use custom name if provided, otherwise auto-generate
            if 'func_name' in kwargs:
                func_name = kwargs['func_name']
            else:
                self.func_counter += 1
                func_name = f"func{self.func_counter}"
            
            # Get parameter names from the property (no flattening needed)
            param_names = getattr(node, 'param_names', [])
            param_str = ', '.join(param_names) if param_names else ''
            
            # Flatten body
            body_stmts, body_expr = self.flatten_ast(body_node)
            
            # Generate function definition
            statements.append(f"def {func_name}({param_str}):")
            if body_stmts:
                # Multi-line function body - optimize like lines 526-536
                last_temp_var = f'temp{self.temp_counter}'
                if body_expr == last_temp_var:
                    # Replace last temp assignment with return (similar to result optimization)
                    for stmt in body_stmts[:-1]:
                        statements.append(f"    {stmt}")
                    statements.append(f"    {body_stmts[-1].replace(f'{last_temp_var} = ', 'return ', 1)}")
                else:
                    # Handle case where body_expr might be a list or different expression
                    for stmt in body_stmts:
                        statements.append(f"    {stmt}")
                    if isinstance(body_expr, list):
                        body_expr = ', '.join(body_expr)
                    statements.append(f"    return {body_expr}")
            else:
                # Single-line function body
                if isinstance(body_expr, list):
                    body_expr = ', '.join(body_expr)
                statements.append(f"    return {body_expr}")
            
            return statements, func_name
                
        elif node.node_type == 'dict':
            # Flatten dictionary
            keys_node = node.children[0]
            values_node = node.children[1]
            k_stmts, k_expr = self.flatten_ast(keys_node)
            v_stmts, v_expr = self.flatten_ast(values_node)
            statements.extend(k_stmts)
            statements.extend(v_stmts)
            return statements, f"{{{k_expr}: {v_expr}}}"
        else:
            return statements, str(node)

# AST Transformer
def transform_ast(node: ASTNode) -> str:
    """Transform AST nodes to Python-like syntax"""
    if node.node_type == 'variable':
        return node.value  # Variables don't get backticks
    elif node.node_type == 'symbol':
        return f"`{node.value}"  # Only symbols get backticks
    elif node.node_type == 'symbol_list':
        # Flatten symbol list into individual symbols
        return [f"`{elem.value}" for elem in node.children]
    elif node.node_type == 'integer':
        return str(node.value)
    elif node.node_type == 'float':
        # Ensure float has decimal point
        value_str = str(node.value)
        if '.' not in value_str:
            value_str += '.0'
        return value_str
    elif node.node_type == 'boolean':
        return 'True' if node.value else 'False'
    elif node.node_type == 'string':
        return node.value
    elif node.node_type == 'function':
        # Convert function calls from LISP [Func[name], arg1, arg2] to name(arg1, arg2)
        # Replace glyphs with readable names
        func_name = glyph_map.get(node.value, node.value)
        
        # Transform arguments with flattening
        arg_strs = []
        for arg in node.children:
            arg_result = transform_ast(arg)
            if isinstance(arg_result, list):
                arg_strs.extend(arg_result)
            else:
                arg_strs.append(arg_result)
        
        args_str = ', '.join(arg_strs)
        return f"{func_name}({args_str})"
    elif node.node_type == 'list':
        # Check if this is a function call (first element is Function)
        if node.children and node.children[0].node_type == 'function':
            func = node.children[0]
            func.children = node.children[1:]  # Set the arguments
            return transform_ast(func)
        else:
            # Regular list
            elem_strs = []
            for elem in node.children:
                elem_result = transform_ast(elem)
                if isinstance(elem_result, list):
                    elem_strs.extend(elem_result)
                else:
                    elem_strs.append(elem_result)
            elements_str = ', '.join(elem_strs)
            return f"[{elements_str}]"
    elif node.node_type == 'dict':
        # Transform Dict to Python dict syntax
        keys_node = node.children[0]
        values_node = node.children[1]
        k_str = transform_ast(keys_node)
        v_str = transform_ast(values_node)
        return f"{{{k_str}: {v_str}}}"
    else:
        return str(node)

def convert_lisp_to_function_calls(parsed_str):
    """Convert LISP-like representation to function call syntax using AST"""
    # First decode bytes to string if needed
    if isinstance(parsed_str, bytes):
        parsed_str = parsed_str.decode('utf-8')
    
    # Tokenize and parse into AST
    tokens = tokenize(parsed_str)
    parser = Parser(tokens)
    ast = parser.parse()
    
    # Transform AST to Python-like syntax
    return transform_ast(ast)

def convert_lisp_to_flat_statements(parsed_str):
    """Convert LISP-like representation to step-by-step Python statements"""
    # First decode bytes to string if needed
    if isinstance(parsed_str, bytes):
        parsed_str = parsed_str.decode('utf-8')
    
    # Tokenize and parse into AST
    tokens = tokenize(parsed_str)
    parser = Parser(tokens)
    ast = parser.parse()
    
    # Flatten AST to step-by-step assignments
    f = Flattener()
    statements, final_expr = f.flatten(ast)
    # Format output
    if statements:
        # Add the final expression without temp variable (unless it's None for assignments)
        if final_expr is not None:
            # Check if final_expr is a temp variable that we can replace with its expression
            if final_expr.startswith('temp') and final_expr[4:].isdigit():
                # Replace the last temp assignment with just the expression
                last_stmt = statements[-1]
                if last_stmt.startswith(f"{final_expr} = "):
                    # Remove the temp assignment and use the expression directly
                    statements[-1] = last_stmt[len(f"{final_expr} = "):]
                else:
                    statements.append(final_expr)
            else:
                statements.append(final_expr)
        return '\n'.join(statements)
    else:
        # No statements, just return the final expression
        return final_expr
    # if statements:
    #     if final_expr == last_temp_var:
    #         statements[-1] = statements[-1].replace(last_temp_var, 'result', 1)
    #     else: 
    #         # Handle case where final_expr might be a list
    #         if isinstance(final_expr, list):
    #             final_expr = ', '.join(final_expr)
    #         # Skip result assignment if it would just be "result = func{n}"
    #         if not (final_expr.startswith('func') and final_expr[4:].isdigit()):
    #             statements.append('result = ' + final_expr)
    # else:
    #     # Handle case where final_expr might be a list
    #     if isinstance(final_expr, list):
    #         final_expr = ', '.join(final_expr)
    #     # Skip result assignment if it would just be "result = func{n}"
    #     if final_expr.startswith('func') and final_expr[4:].isdigit():
    #         return final_expr
    #     else:
    #         return 'result = ' + final_expr

def translate(q_expression: str, q) -> str:
    """
    Take a q expression, convert to flattened Python-like code,
    then ask Claude to add disambiguating comments
    """
    # Convert q to flattened Python-like code
    if q._current_async_task and q._current_async_task.get("status") == "Running":
        return "Cannot use code q->qython code translation while another q query is running"
    flattened_code = convert_lisp_to_flat_statements(parseq0(q_expression, q))
    
    return disambiguate_step(flattened_code)


def parseq0(s, qcomms):
    assert len(qcomms._listener_thread.messages) == 0 , "Shouldn't happen: there are messages in the listener thread message stack upon calling parseq0 in parseq module"
    qcomms._q_connection.q.sendSyncDetached(parseq_q_code+'; .parseq.parseq', s)
    assert qcomms._listener_thread.wait_for_response(1) # wait for maximum 1 second though in practice should be instant
    return qcomms.parse_raw_bytes(qcomms._listener_thread.messages.pop().data).str

def parseq1(s, q):
    return convert_lisp_to_function_calls(parseq0(s, q))

def parseq(s, q):
    # return convert_lisp_to_function_calls(q('var2string parse "'+s+'"'))
    return convert_lisp_to_flat_statements(parseq0(s, q))
def pr(s, q):
    print(parseq(s, q),'\n')
# pr('f[min]')
# pr('a lj 2!select min s, maxs t from c')
# pr('min x')
# pr('f x')
# pr('var')
# pr('`var')
# pr('`a`s`d')
# pr('f[a;b;c]')
# pr('f[`a`b`c]')
# pr('2+3*4')
# pr('{x+1}')
# print(parseq0('{a:x;x+a}'))
# pr('{a:x;x+a}')
# print(parseq0('nmsq:{[x] ({(y+x%y)%2f}[x]/) x%2f}'),'\n')
# print(parseq1('nmsq:{[x] ({(y+x%y)%2f}[x]/) x%2f}'),'\n')
# print(parseq('nmsq:{[x] ({(y+x%y)%2f}[x]/) x%2f}'),'\n')
# pr('f1:{x+1};f2:{x+2}')
# pr('2+3')




"""
Utility functions for qmcp server
"""

import psutil


def getTimeoutsStr(_switch_to_async_timeout, _interrupt_timeout, _connection_timeout):
    """Format timeout settings as a readable string"""
    async_setting = f"{_switch_to_async_timeout}s" if _switch_to_async_timeout else "disabled"
    interrupt_setting = f"{_interrupt_timeout}s" if _interrupt_timeout else "disabled"
    connection_setting = f"{_connection_timeout}s"
    return f"Timeouts: async_switch={async_setting}, interrupt={interrupt_setting}, connection={connection_setting}"


def find_process_by_port(port):
    """Find PID of process listening on the specified port"""
    if not port:
        return None
    try:
        for conn in psutil.net_connections():
            if (conn.laddr.port == port and 
                conn.status == 'LISTEN'):
                return conn.pid
    except Exception:
        pass
    return None


class MetaData(object):
    '''Utility class for enriching data structures with meta data, e.g. qtype hint.'''
    def __init__(self, **kw):
        self.__dict__.update(kw)

    def __repr__(self):
        if not self.__dict__.items():
            return 'metadata()'

        s = ['metadata(']
        for k, v in self.__dict__.items():
            s.append('%s=%s' % (k, repr(v)))
            s.append(', ')
        s[-1] = ')'
        return ''.join(s)

    def __getattr__(self, attr):
        return None

    def __getitem__(self, key):
        return self.__dict__.get(key, None)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def as_dict(self):
        return self.__dict__.copy()

    def union_dict(self, **kw):
        return dict(list(self.as_dict().items()) + list(kw.items()))


CONVERSION_OPTIONS = MetaData(raw = False,
                              numpy_temporals = False,
                              pandas = False,
                              single_char_strings = False
                             )